import os
import shutil
from embit import bip39, bip32, networks
from . import DeviceTypes
from ..device import Device
from ..helpers import alias
from ..util.descriptor import AddChecksum
from ..util.base58 import encode_base58_checksum, decode_base58
from ..util.xpub import get_xpub_fingerprint, convert_xpub_prefix
from ..key import Key
from ..rpc import get_default_datadir
from io import BytesIO
import hmac
import logging

logger = logging.getLogger(__name__)


class BitcoinCore(Device):
    device_type = DeviceTypes.BITCOINCORE
    name = "Bitcoin Core (hot wallet)"
    icon = "bitcoincore_icon.svg"

    hot_wallet = True
    taproot_support = True

    def __init__(self, *args, **kwargs):
        self._use_descriptors = None
        super().__init__(*args, **kwargs)

    def setup_device(self, file_password, wallet_manager):
        wallet_name = os.path.join(wallet_manager.rpc_path + "_hotstorage", self.alias)
        core_version = wallet_manager.rpc.getnetworkinfo().get("version", 0)
        use_descriptors = core_version >= 210000
        if use_descriptors:
            # it could fail if Core is compiled without sqlite
            try:
                wallet_manager.rpc.createwallet(
                    wallet_name, False, True, file_password or "", False, True
                )
                self._use_descriptors = True
            except:
                wallet_manager.rpc.createwallet(
                    wallet_name, False, True, file_password or "", False, True
                )
                self.taproot_support = False
        else:
            wallet_manager.rpc.createwallet(
                wallet_name, False, True, file_password or ""
            )
        rpc = wallet_manager.rpc.wallet(wallet_name)

    def use_descriptors(self, rpc):
        if self._use_descriptors is None:
            self._use_descriptors = rpc.getwalletinfo().get("descriptors", False)
            if not self._use_descriptors:
                self.taproot_support = False
        return self._use_descriptors

    def taproot_available(self, rpc):
        try:
            # currently only master branch supports tr() descriptors
            # TODO: replace to 220000
            core_version = rpc.getnetworkinfo().get("version", 0)
            info = rpc.getblockchaininfo()
            taproot_active = (core_version >= 219900) and (
                info.get("softforks", {}).get("taproot", {}).get("active", False)
            )
            taproot_support = self.use_descriptors(rpc) and taproot_active
            self.taproot_support = taproot_support
            return taproot_support
        except Exception as e:
            self.taproot_support = False
            logger.error(e)
            return False

    def add_hot_wallet_keys(
        self,
        mnemonic,
        passphrase,
        paths,
        file_password,
        wallet_manager,
        testnet,
        keys_range=[0, 1000],
        keys_purposes=[],
    ):
        if type(keys_range[0]) == str:
            keys_range[0] = int(keys_range[0])
        if type(keys_range[1]) == str:
            keys_range[1] = int(keys_range[1])
        seed = bip39.mnemonic_to_seed(mnemonic, passphrase)
        root = bip32.HDKey.from_seed(seed)
        network = networks.NETWORKS["test" if testnet else "main"]
        root.version = network["xprv"]
        xprv = root.to_base58()
        # Load the wallet if not loaded
        self._load_wallet(wallet_manager)
        rpc = wallet_manager.rpc.wallet(
            os.path.join(wallet_manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)

        # check if we can use descriptors and taproot
        use_descriptors = self.use_descriptors(rpc)
        taproot_available = self.taproot_available(rpc)

        args = [
            {
                "desc": AddChecksum(
                    "{}({}{}/0/*)".format(
                        "tr"
                        if path.startswith("m/86h") and taproot_available
                        else "wpkh",
                        xprv,
                        path.rstrip("/").replace("m", ""),
                    )
                ),
                "range": keys_range,
                "timestamp": "now",
            }
            for path in paths
        ] + [
            {
                "desc": AddChecksum(
                    "{}({}{}/1/*)".format(
                        "tr"
                        if path.startswith("m/86h") and taproot_available
                        else "wpkh",
                        xprv,
                        path.rstrip("/").replace("m", ""),
                    )
                ),
                "range": keys_range,
                "timestamp": "now",
            }
            for path in paths
        ]

        if use_descriptors:
            rpc.importdescriptors(args)
        else:
            rpc.importmulti(args, {"rescan": False})

        xpubs = [root.derive(path).to_public().to_base58() for path in paths]
        # root fingerprint is fingerprint field of the first child
        master_fpr = root.child(0).fingerprint.hex()
        keys = []
        for i in range(len(paths)):
            try:
                path = paths[i]
                xpub = xpubs[i]
                # detect slip132 version for xpubs
                slip132_prefix = bip32.detect_version(
                    path, default="xpub", network=network
                )
                xpub = "[{}{}]{}\n".format(
                    master_fpr,
                    path.replace("m", ""),
                    convert_xpub_prefix(xpub, slip132_prefix),
                )
                keys.append(
                    Key.parse_xpub(
                        xpub, keys_purposes[i] if len(keys_purposes) > i else ""
                    )
                )
            except Exception:
                # TODO: This should never occur, but just in case,
                # we must make sure to catch it properly so it
                # doesn't crash the app no matter what.
                raise Exception("Failed to parse this xpub:\n" + "\n".join(xpub))
        self.add_keys(keys)

    def _load_wallet(self, wallet_manager):
        try:
            existing_wallets = [
                w["name"] for w in wallet_manager.rpc.listwalletdir()["wallets"]
            ]
        except:
            existing_wallets = None
        loaded_wallets = wallet_manager.rpc.listwallets()

        hotstorage_path = wallet_manager.rpc_path + "_hotstorage"
        wallet_path = os.path.join(hotstorage_path, self.alias)
        if existing_wallets is None or wallet_path in existing_wallets:
            if wallet_path not in loaded_wallets:
                wallet_manager.rpc.loadwallet(wallet_path)

    def is_encrypted(self, wallet_manager):
        """Check if the wallet is encrypted"""
        self._load_wallet(wallet_manager)
        hotstorage_path = wallet_manager.rpc_path + "_hotstorage"
        wallet_path = os.path.join(hotstorage_path, self.alias)
        try:
            # check if password is enabled
            info = wallet_manager.rpc.getwalletinfo(wallet=wallet_path)
            return "unlocked_until" in info
        except Exception as e:
            logger.warning("Cannot fetch hot wallet info")
        # Assuming encrypted by default
        return True

    def create_psbts(self, base64_psbt, wallet):
        return {"core": base64_psbt}

    def sign_psbt(self, base64_psbt, wallet, file_password=None):
        # Load the wallet if not loaded
        self._load_wallet(wallet.manager)
        rpc = wallet.manager.rpc.wallet(
            os.path.join(wallet.manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)
        signed_psbt = rpc.walletprocesspsbt(base64_psbt)
        if base64_psbt == signed_psbt["psbt"]:
            raise Exception(
                "Make sure you have entered the wallet file password correctly. (If your wallet is not encrypted submit empty password)"
            )
        if file_password:
            rpc.walletlock()
        return signed_psbt

    def sign_raw_tx(self, raw_tx, wallet, file_password=None):
        # Load the wallet if not loaded
        self._load_wallet(wallet.manager)
        rpc = wallet.manager.rpc.wallet(
            os.path.join(wallet.manager.rpc_path + "_hotstorage", self.alias)
        )
        if file_password:
            rpc.walletpassphrase(file_password, 60)
        signed_tx = rpc.signrawtransactionwithwallet(raw_tx)
        if file_password:
            rpc.walletlock()
        return signed_tx

    def delete(
        self, wallet_manager, bitcoin_datadir=get_default_datadir(), chain="main"
    ):
        try:
            wallet_rpc_path = os.path.join(
                wallet_manager.rpc_path + "_hotstorage", self.alias
            )
            wallet_manager.rpc.unloadwallet(wallet_rpc_path)
            # Try deleting wallet file
            if bitcoin_datadir:
                if chain != "main":
                    bitcoin_datadir = os.path.join(bitcoin_datadir, chain)
                candidates = [
                    os.path.join(bitcoin_datadir, wallet_rpc_path),
                    os.path.join(bitcoin_datadir, "wallets", wallet_rpc_path),
                ]
                for path in candidates:
                    if os.path.exists(path):
                        shutil.rmtree(path)
                        break
        except:
            pass  # We tried...



import logging

from .base import (
    SecureResource,
    rest_resource,
)
from flask import current_app as app, request
from ...wallet import Wallet
from ...util.psbt_creator import PsbtCreator

from .. import auth

logger = logging.getLogger(__name__)


@rest_resource
class ResourcePsbt(SecureResource):
    """/api/v1alpha/specter"""

    endpoints = ["/v1alpha/wallets/<wallet_alias>/psbt"]

    def get(self, wallet_alias):
        # ToDo: check whether the user has access to the wallet
        user = auth.current_user()
        wallet: Wallet = app.specter.user_manager.get_user(
            user
        ).wallet_manager.get_by_alias(wallet_alias)
        pending_psbts = wallet.pending_psbts
        return {"result": pending_psbts or []}

    def post(self, wallet_alias):
        user = auth.current_user()
        wallet: Wallet = app.specter.user_manager.get_user(
            user
        ).wallet_manager.get_by_alias(wallet_alias)
        logger.debug(f"Got a post request for creating a psbt: {request.json}")
        psbt_creator = PsbtCreator(
            app.specter, wallet, "json", request_json=request.json
        )
        psbt_creator.create_psbt(wallet)
        return {"result": psbt_creator.psbt}



from embit.liquid.pset import PSET


def to_canonical_pset(pset):
    """
    Removes unblinded information from the transaction
    so Elements Core can decode it
    """
    # if we got psbt, not pset - just return
    if not pset.startswith("cHNl"):
        return pset
    tx = PSET.from_string(pset)

    for inp in tx.inputs:
        inp.value = None
        inp.asset = None
        inp.value_blinding_factor = None
        inp.asset_blinding_factor = None

    for out in tx.outputs:
        if out.is_blinded:
            out.asset = None
            out.asset_blinding_factor = None
            out.value = None
            out.value_blinding_factor = None
    return str(tx)



import os
from embit import bip39
from embit.liquid.slip77 import master_blinding_from_seed

from ..helpers import is_liquid
from . import DeviceTypes
from .bitcoin_core import BitcoinCore


class ElementsCore(BitcoinCore):
    device_type = DeviceTypes.ELEMENTSCORE
    name = "Elements Core (hot wallet)"
    icon = "elementscore_icon.svg"

    hot_wallet = True
    bitcoin_core_support = False
    liquid_support = True

    def add_hot_wallet_keys(
        self,
        mnemonic,
        passphrase,
        paths,
        file_password,
        wallet_manager,
        testnet,
        keys_range=[0, 1000],
        keys_purposes=[],
    ):
        super().add_hot_wallet_keys(
            mnemonic,
            passphrase,
            paths,
            file_password,
            wallet_manager,
            testnet,
            keys_range=keys_range,
            keys_purposes=keys_purposes,
        )
        rpc = wallet_manager.rpc.wallet(
            os.path.join(wallet_manager.rpc_path + "_hotstorage", self.alias)
        )
        seed = bip39.mnemonic_to_seed(mnemonic, passphrase)
        master_blinding_key = master_blinding_from_seed(seed)
        rpc.importmasterblindingkey(master_blinding_key.secret.hex())
        self.set_blinding_key(master_blinding_key.wif())

    def has_key_types(self, wallet_type, network="main"):
        if not is_liquid(network):
            return False
        return super().has_key_types(wallet_type, network)

    def no_key_found_reason(self, wallet_type, network="main"):
        if self.has_key_types(wallet_type, network=network):
            return ""
        if not is_liquid(network):
            return "This wallet can only sign on Liquid"



import atexit
import json
import logging
import os
import shutil
import signal
import sys
import time
from pathlib import Path

import click
import psutil
from flask import Config

from ..config import DEFAULT_CONFIG
from ..process_controller.node_controller import find_node_executable
from ..process_controller.elementsd_controller import ElementsPlainController
from .utils import (
    Echo,
    kill_node_process,
    purge_node_data_dir,
    compute_data_dir_and_set_config_obj,
)

logger = logging.getLogger(__name__)

# ------------------------bitcoind -----------------------------------------------------


@click.command()
@click.option("--quiet/--no-quiet", default=False, help="Output as little as possible.")
@click.option(
    "--nodocker",
    default=False,
    is_flag=True,
    help="Use without docker. (By default docker is used.)",
)
@click.option(
    "--docker-tag", "docker_tag", default="latest", help="Use a specific docker-tag"
)
@click.option(
    "--data-dir",
    help="Specify a (maybe not yet existing) datadir. Works only with --nodocker. (Default is /tmp/bitcoind_plain_datadir)",
)
@click.option(
    "--log-stdout",
    is_flag=True,
    default=False,
    help="Will spit out logs on stdout if set",
)
@click.option(
    "--mining/--no-mining",
    default=True,
    help="Turns on mining (On by default). For testing it is useful to turn it off.",
)
@click.option(
    "--mining-period",
    default="15",
    help="Specify mining period (in seconds). Every N seconds a block gets mined. (Default is 15sec)",
)
@click.option(
    "--reset",
    is_flag=True,
    default=False,
    help="Kill the bitcoind daemon. Datadir will get lost.",
)
@click.option(
    "--create-conn-json",
    is_flag=True,
    default=False,
    help="Create a small json-file named btcd-conn.json with connection details.",
)
@click.option(
    "--cleanuphard/--no-cleanuphard",
    default=False,
    help="Send signal SIGKILL instead of SIGTERM (default) when CTRL-C is pressed. Mostly to speed up tests.",
)
@click.option(
    "--config",
    default=None,
    help="A class from the config.py which sets reasonable default values.",
)
def bitcoind(
    quiet,
    nodocker,
    docker_tag,
    data_dir,
    log_stdout,
    mining,
    mining_period,
    reset,
    create_conn_json,
    cleanuphard,
    config,
):
    """This will start a bitcoind regtest and mines a block every mining-period.
    If a bitcoind is already running on port 18443, it won't start another one. If you CTRL-C this, the bitcoind will
    still continue to run. You have to shut it down.
    """
    noded(
        "bitcoin",
        quiet,
        nodocker,
        docker_tag,
        data_dir,
        log_stdout,
        mining,
        mining_period,
        reset,
        create_conn_json,
        cleanuphard,
        config,
    )


# ------------------------elementsd -----------------------------------------------------


@click.command()
@click.option("--quiet/--no-quiet", default=False, help="Output as little as possible.")
@click.option(
    "--data-dir",
    help="Specify a (maybe not yet existing) datadir. Works only with --nodocker. (Default is /tmp/bitcoind_plain_datadir)",
)
@click.option(
    "--log-stdout",
    is_flag=True,
    default=False,
    help="Will spit out logs on stdout if set",
)
@click.option(
    "--mining/--no-mining",
    default=True,
    help="Turns on mining (On by default). For testing it is useful to turn it off.",
)
@click.option(
    "--mining-period",
    default="15",
    help="Specify mining period (in seconds). Every N seconds a block gets mined. (Default is 15sec)",
)
@click.option(
    "--reset",
    is_flag=True,
    default=False,
    help="Kill the bitcoind daemon. Datadir will get lost.",
)
@click.option(
    "--create-conn-json",
    is_flag=True,
    default=False,
    help="Create a small json-file named btcd-conn.json with connection details.",
)
@click.option(
    "--cleanuphard/--no-cleanuphard",
    default=False,
    help="Send signal SIGKILL instead of SIGTERM (default) when CTRL-C is pressed. Mostly to speed up tests.",
)
@click.option(
    "--config",
    default=None,
    help="A class from the config.py which sets reasonable default values.",
)
def elementsd(
    quiet,
    data_dir,
    log_stdout,
    mining,
    mining_period,
    reset,
    create_conn_json,
    cleanuphard,
    config,
):
    """This will start a bitcoind regtest and mines a block every mining-period.
    If a bitcoind is already running on port 18443, it won't start another one. If you CTRL-C this, the bitcoind will
    still continue to run. You have to shut it down.
    """
    noded(
        "elements",
        quiet,
        True,  # nodocker
        None,  # docker_tag
        data_dir,
        log_stdout,
        mining,
        mining_period,
        reset,
        create_conn_json,
        cleanuphard,
        config,
    )


def noded(
    node_impl,
    quiet,
    nodocker,
    docker_tag,
    data_dir,
    log_stdout,
    mining,
    mining_period,
    reset,
    create_conn_json,
    cleanuphard,
    config,
):
    """This will start a noded (depending on node_impl) regtest and mines a block every mining-period.
    If a node_impl is already running on port ?, it won't start another one. If you CTRL-C this, the node_impl will
    still continue to run. You have to shut it down yourself.
    """
    logger.debug("If you can read, this, logging is on debug")
    # In order to avoid these dependencies for production use, we're importing them here:
    import docker

    from ..process_controller.bitcoind_controller import BitcoindPlainController

    if config is None:
        config = DEFAULT_CONFIG
    else:
        if not "." in config:
            config = "cryptoadvance.specter.config." + config
    config_obj = Config(".")
    config_obj.from_object(config)

    echo = Echo(quiet).echo

    data_dir = compute_data_dir_and_set_config_obj(node_impl, data_dir, config_obj)

    if reset:
        if not nodocker:
            echo("ERROR: --reset only works in conjunction with --nodocker currently")
            return
        did_something = kill_node_process(node_impl, echo)
        did_something = (
            purge_node_data_dir(node_impl, config_obj, echo) or did_something
        )
        if not did_something:
            echo("Nothing to do!")
        return
    mining_every_x_seconds = float(mining_period)
    if nodocker:
        echo(f"Creating plain {node_impl}d")
        if node_impl == "bitcoin":
            my_node = BitcoindPlainController(
                bitcoind_path=find_node_executable("bitcoin")
            )
        elif node_impl == "elements":
            my_node = ElementsPlainController(
                elementsd_path=find_node_executable("elements")
            )
        Path(data_dir).mkdir(parents=True, exist_ok=True)
    else:
        echo("Creating container")
        from ..process_controller.bitcoind_docker_controller import (
            BitcoindDockerController,
        )

        if node_impl == "bitcoin":
            my_node = BitcoindDockerController(docker_tag=docker_tag)
        else:
            raise Exception("There is no Elementsd-Bitcoin-Controller yet!")
    try:
        echo(f"Starting {node_impl}d")
        if node_impl == "bitcoin":
            my_node.start_bitcoind(
                cleanup_at_exit=True,
                cleanup_hard=cleanuphard,
                datadir=data_dir,
                log_stdout=log_stdout,
            )
        elif node_impl == "elements":
            my_node.start_elementsd(
                cleanup_at_exit=True,
                cleanup_hard=cleanuphard,
                datadir=data_dir,
                log_stdout=log_stdout,
            )
    except docker.errors.ImageNotFound:
        echo(f"Image with tag {docker_tag} does not exist!")
        echo(
            f"Try to download first with docker pull registry.gitlab.com/cryptoadvance/specter-desktop/python-bitcoind:{docker_tag}"
        )
        sys.exit(1)
    except Exception as e:
        if str(e).startswith("There is already a node running!"):
            echo(f"{e} please reset via:")
            echo(f"python3 -m cryptoadvance.specter {node_impl}d --reset")
        else:
            raise e
    if not nodocker:
        tags_of_image = [
            image.split(":")[-1] for image in my_node.btcd_container.image.tags
        ]
        if docker_tag not in tags_of_image:
            echo(
                "The running docker container is not \
                                the tag you requested!"
            )
            echo(
                "please stop first with docker stop {}".format(
                    my_node.btcd_container.id
                )
            )
            sys.exit(1)
        echo(f"containerImage: {my_node.btcd_container.image.tags} ")
    echo(f"           url: {my_node.rpcconn.render_url()}")
    echo(f"user, password: { my_node.rpcconn.rpcuser }, secret")
    echo(f"    host, port: localhost, {my_node.rpcconn.rpcport}")
    echo(
        f"   {node_impl}-cli: {node_impl}-cli -regtest -rpcport={my_node.rpcconn.rpcport} -rpcuser={ my_node.rpcconn.rpcuser } -rpcpassword=secret getblockchaininfo "
    )

    if create_conn_json:
        conn = my_node.rpcconn.as_data()
        conn["pid"] = os.getpid()  # usefull to send signals
        conn["specter_data_folder"] = config_obj[
            "SPECTER_DATA_FOLDER"
        ]  # e.g. cypress might want to know where we're mining to
        conn[f"{node_impl}_data_dir"] = data_dir
        conn_file = f"{'btcd' if node_impl == 'bitcoin' else 'elmd'}-conn.json"
        with open(conn_file, "w") as file:
            file.write(json.dumps(conn))

        def cleanup():
            os.remove(conn_file)

        atexit.register(cleanup)

    signal.signal(
        signal.SIGUSR1,
        lambda x, y: mine_2_specter_wallets(
            node_impl, my_node, config_obj["SPECTER_DATA_FOLDER"], echo
        ),
    )

    if node_impl == "elements":
        prepare_elements_default_wallet(my_node)

    if mining:
        miner_loop(
            node_impl,
            my_node,
            config_obj["SPECTER_DATA_FOLDER"],
            mining_every_x_seconds,
            echo,
        )


def prepare_elements_default_wallet(my_node):
    """this will collect the free coins we have created with -initialfreecoins=2100000000000000
    and transfer them to the default-wallet
    """
    rpc = my_node.rpcconn.get_rpc()
    wallet = rpc.wallet("")
    freehash = rpc.getblockhash(0)
    freetxid = rpc.getblock(freehash)["tx"][1]
    logger.debug(f"freetxid: {freetxid}")
    if rpc.gettxout(freetxid, 0):  # unspent!
        tx = rpc.getrawtransaction(freetxid, 1)
        fee = 1000e-8
        value = round(tx["vout"][0]["value"] - fee, 8)
        addr = wallet.getnewaddress()
        unconfidential = wallet.getaddressinfo(addr)["unconfidential"]
        rawtx = wallet.createrawtransaction(
            [{"txid": freetxid, "vout": 0}],  # inputs
            [{unconfidential: value}, {"fee": fee}],
        )
        wallet.sendrawtransaction(rawtx)
        rpc.generatetoaddress(101, unconfidential)


def miner_loop(node_impl, my_node, data_folder, mining_every_x_seconds, echo):
    "An endless loop mining bitcoin"

    echo(
        "Now, mining a block every %f seconds, avoid it via --no-mining"
        % mining_every_x_seconds
    )
    mine_2_specter_wallets(node_impl, my_node, data_folder, echo)

    # make them spendable
    my_node.mine(block_count=100)
    echo(
        f"height: {my_node.rpcconn.get_rpc().getblockchaininfo()['blocks']} | ",
        nl=False,
    )
    i = 0
    while True:
        try:
            my_node.mine()
            current_height = my_node.rpcconn.get_rpc().getblockchaininfo()["blocks"]
        except Exception as e:
            logger.debug(
                f"Caught {e}, Couldn't mine, assume SIGTERM occured => exiting!"
            )
            echo(f"THE_END(@height:{current_height})")
            break
        echo("%i" % (i % 10), prefix=False, nl=False)
        if i % 10 == 9:
            echo(" ", prefix=False, nl=False)
        i += 1
        if i >= 50:
            i = 0
            echo("", prefix=False)
            echo(
                f"height: {current_height} | ",
                nl=False,
            )
        time.sleep(mining_every_x_seconds)


def mine_2_specter_wallets(node_impl, my_node, data_folder, echo):
    """Get each specter-wallet some coins"""

    from ..process_controller.node_controller import fetch_wallet_addresses_for_mining

    try:

        for address in fetch_wallet_addresses_for_mining(node_impl, data_folder):
            echo("")
            echo(f"Mining to address {address}")
            my_node.testcoin_faucet(address)
            # my_node.mine(address=address)
        # my_node.mine(block_count=100)
    except FileNotFoundError:
        # might happen if there no ~/.specter folder yet
        pass



""" Stuff to control a bitcoind-instance.
"""
import atexit
import json
import logging
import os
import platform
import re
import shutil
import signal
import subprocess
import tempfile
import time

import psutil
from cryptoadvance.specter.liquid.rpc import LiquidRPC
from requests.exceptions import ConnectionError
from urllib3.exceptions import MaxRetryError, NewConnectionError

from ..helpers import load_jsons
from ..rpc import BitcoinRPC, RpcError
from ..specter_error import ExtProcTimeoutException, SpecterError
from ..util.shell import get_last_lines_from_file, which

logger = logging.getLogger(__name__)


class Btcd_conn:
    """An object to easily store connection data to bitcoind-comatible nodes (Bitcoin/Elements)"""

    def __init__(
        self,
        node_impl="bitcoin",
        rpcuser="bitcoin",
        rpcpassword="secret",
        rpcport=18543,
        ipaddress=None,
    ):
        self.node_impl = node_impl
        self.rpcport = rpcport
        self.rpcuser = rpcuser
        self.rpcpassword = rpcpassword
        self._ipaddress = ipaddress

    @property
    def ipaddress(self):
        if self._ipaddress == None:
            raise Exception("ipadress is none")
        else:
            return self._ipaddress

    @ipaddress.setter
    def ipaddress(self, ipaddress):
        self._ipaddress = ipaddress

    def get_rpc(self):
        """returns a BitcoinRPC"""
        if self.node_impl == "bitcoin":
            rpc = BitcoinRPC(
                self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport
            )
        elif self.node_impl == "elements":
            rpc = LiquidRPC(
                self.rpcuser, self.rpcpassword, host=self.ipaddress, port=self.rpcport
            )
        else:
            raise SpecterError(f"Unknown node_impl: {self.node_impl}")
        rpc.getblockchaininfo()
        return rpc

    def render_url(self, password_mask=False):
        if password_mask == True:
            password = "xxxxxx"
        else:
            password = self.rpcpassword
        return "http://{}:{}@{}:{}/wallet/".format(
            self.rpcuser, password, self.ipaddress, self.rpcport
        )

    def as_data(self):
        """returns a data-representation of this connection"""
        me = {
            "user": self.rpcuser,
            "password": self.rpcpassword,
            "host": self.ipaddress,
            "port": self.rpcport,
            "url": self.render_url(),
        }
        return me

    def render_json(self):
        """returns a json-representation of this connection"""
        return json.dumps(self.as_data())

    def __repr__(self):
        return "<Btcd_conn {}>".format(self.render_url())


class NodeController:
    """A kind of abstract class to simplify running a bitcoind with or without docker"""

    def __init__(
        self,
        node_impl,
        rpcport=18443,
        network="regtest",
        rpcuser="bitcoin",
        rpcpassword="secret",
    ):
        try:
            self.rpcconn = Btcd_conn(
                node_impl=node_impl,
                rpcuser=rpcuser,
                rpcpassword=rpcpassword,
                rpcport=rpcport,
            )
            self.network = network
            self.node_impl = node_impl
            # reasonable default
            self.cleanup_hard = False
        except Exception as e:
            logger.exception(f"Failed to instantiate BitcoindController. Error: {e}")
            raise e

    def start_node(
        self,
        cleanup_at_exit=False,
        cleanup_hard=False,
        datadir=None,
        log_stdout=False,
        extra_args=[],
        timeout=60,
    ):
        """starts bitcoind with a specific rpcport=18543 by default.
        That's not the standard in order to make pytest running while
        developing locally against a different regtest-instance
        if bitcoind_path == docker, it'll run bitcoind via docker.
        Specify a longer timeout for slower devices (e.g. Raspberry Pi)
        """
        if not self.check_existing() is None:
            # This should not happen. For bitcoind, have a look in BitcoindController.attach_to_proc_id()
            raise SpecterError(
                f"While starting Node, there is already a Node running at {self.rpcconn.render_url(password_mask=True)}"
            )

        self._start_node(
            cleanup_at_exit,
            cleanup_hard=cleanup_hard,
            datadir=datadir,
            log_stdout=log_stdout,
            extra_args=extra_args,
        )
        try:
            self.wait_for_node(self.rpcconn, timeout=timeout)
            self.status = "Running"
        except ExtProcTimeoutException as e:
            if self.node_proc.poll() is None:
                self.status = "error"
                raise Exception(f"Could not start node due to: {self.get_debug_log()}")
            self.status = "Starting up"
            logger.error(self.node_proc.stdout)
            raise e
        except Exception as e:
            self.status = "Error"
            raise e
        logger.info(f"Successfully started {self.node_impl}d in {self.datadir}")
        if "" not in self.get_rpc().listwallets():
            logger.info("Creating Default-wallet")
            self.get_rpc().createwallet("", False, False, "", False, True, True)

        if self.network == "regtest":
            logger.debug(f"Mining 100 blocks (network: {self.network})")
            self.mine(block_count=100)

        return self.rpcconn

    def version(self):
        raise Exception("version needs to be overridden by Subclassses")

    def is_testnet(self):
        return self.network not in [
            "mainnet",
            "main",
            "liquidv1",
            "None",
            "none",
            None,
            "",
        ]

    def is_liquid(self):
        return self.network not in [
            "main",
            "mainnet",
            "regtest",
            "test",
            "signet",
            "None",
            "none",
            None,
            "",
        ]

    def get_rpc(self):
        """wrapper for convenience"""
        return self.rpcconn.get_rpc()

    def _start_node(
        self,
        cleanup_at_exit,
        cleanup_hard=False,
        datadir=None,
        log_stdout=False,
        extra_args=[],
    ):
        raise Exception(f"This should not be used in the baseclass! self: {self}")

    def get_debug_log(self):
        raise Exception(f"This should not be used in the baseclass! self: {self}")

    def check_existing(self):
        raise Exception(f"This should not be used in the baseclass! self: {self}")

    def stop_node(self):
        raise Exception(f"This should not be used in the baseclass! self: {self}")

    def mine(self, address=None, block_count=1):
        """Does mining to the attached address with as many as block_count blocks"""
        if address == None:
            if self.node_impl == "bitcoin":
                address = "mruae2834buqxk77oaVpephnA5ZAxNNJ1r"
            else:
                address = "el1qqf6tv4n8qp55qc04v4xts5snd9v5uurkry4vskef6lmecahj6c42jt9lnj0432287rs67z9vzq2zvuer036s5mahptwxgyd8k"
        self.get_rpc().generatetoaddress(block_count, address)

    def testcoin_faucet(self, address, amount=20, confirm_payment=True):
        """an easy way to get some testcoins"""
        rpc = self.get_rpc()
        try:
            default_rpc = rpc.wallet("")
            default_rpc.getbalance()
        except RpcError as rpce:
            # return-codes:
            # https://github.com/bitcoin/bitcoin/blob/v0.15.0.1/src/rpc/protocol.h#L32L87
            if rpce.error_code == -18:  # RPC_WALLET_NOT_FOUND
                logger.debug("Creating default wallet")
                rpc.createwallet("")
                default_rpc = rpc.wallet("")
            else:
                raise rpce
        balance = default_rpc.getbalance()
        if isinstance(balance, dict):
            balance = balance["bitcoin"]  # elements
        logger.debug("balance:" + str(balance))
        default_address = default_rpc.getnewaddress("")
        if self.node_impl == "elements":
            default_address = default_rpc.getaddressinfo(default_address)[
                "unconfidential"
            ]
        while True:
            btc_balance = default_rpc.getbalance()
            rpc.generatetoaddress(102, default_address)
            if btc_balance > amount:
                break
        default_rpc.sendtoaddress(address, amount)
        if confirm_payment:
            # confirm it
            rpc.generatetoaddress(1, default_address)

    @staticmethod
    def check_node(rpcconn, raise_exception=False):
        """returns true if bitcoind is running on that address/port"""
        if raise_exception:
            rpcconn.get_rpc()  # that call will also check the connection
            return True
        try:
            rpcconn.get_rpc()  # that call will also check the connection
            return True
        except RpcError as e:
            # E.g. "Loading Index ..." #ToDo: check it here
            return False
        except ConnectionError as e:
            return False
        except MaxRetryError as e:
            return False
        except TypeError as e:
            return False
        except NewConnectionError as e:
            return False
        except Exception as e:
            # We should avoid this:
            # If you see it in the logs, catch that new exception above
            logger.error("Unexpected Exception, THIS SHOULD NOT HAPPEN " + str(type(e)))
            logger.debug(f"could not reach bitcoind - message returned: {e}")
            return False

    @staticmethod
    def wait_for_node(rpcconn, timeout=15):
        """tries to reach the bitcoind via rpc. Timeout after n seconds"""
        i = 0
        while True:
            if NodeController.check_node(rpcconn):
                break
            time.sleep(0.5)
            i = i + 1
            if i % 10 == 0:
                logger.info(f"Node timeout in {timeout - i/2} seconds")
            if i > (2 * timeout):
                try:
                    NodeController.check_node(rpcconn, raise_exception=True)
                except Exception as e:
                    if "Verifying blocks..." in str(e) or "Loading wallet..." in str(e):
                        # Timed out too soon while node is still spinning up
                        logger.info("Giving node more time to restart...")
                        i = 0
                        pass
                    else:
                        raise ExtProcTimeoutException(
                            f"Timeout while trying to reach node {rpcconn.render_url(password_mask=True)} because {e}".format(
                                rpcconn
                            )
                        )

    @staticmethod
    def render_rpc_options(rpcconn):
        options = " -rpcport={} -rpcuser={} -rpcpassword={} ".format(
            rpcconn.rpcport, rpcconn.rpcuser, rpcconn.rpcpassword
        )
        return options

    @classmethod
    def construct_node_cmd(
        cls,
        rpcconn,
        run_docker=True,
        datadir=None,
        log_stdout=False,
        node_path="bitcoind",
        network="regtest",
        extra_args=[],
    ):
        """returns a command to run your node (bitcoind/elementsd)"""
        btcd_cmd = '"{}" '.format(node_path)
        if network != "mainnet" and network != "main":
            btcd_cmd += " -{} ".format(network)
        btcd_cmd += " -fallbackfee=0.0002 "
        btcd_cmd += " -port={} -rpcport={} -rpcbind=0.0.0.0 -rpcbind=0.0.0.0".format(
            rpcconn.rpcport - 1, rpcconn.rpcport
        )
        btcd_cmd += " -rpcuser={} -rpcpassword={} ".format(
            rpcconn.rpcuser, rpcconn.rpcpassword
        )
        btcd_cmd += " -rpcallowip=0.0.0.0/0 -rpcallowip=172.17.0.0/16 "
        if not run_docker:
            if not log_stdout:
                btcd_cmd += " -noprinttoconsole"
            if datadir == None:
                datadir = tempfile.mkdtemp(prefix="bitcoind_datadir")
            btcd_cmd += ' -datadir="{}" '.format(datadir)
        if extra_args:
            btcd_cmd += " {}".format(" ".join(extra_args))
        logger.debug("constructed bitcoind-command: %s", btcd_cmd)
        return btcd_cmd


class NodePlainController(NodeController):
    """A class controlling the bitcoind-process directly on the machine"""

    def __init__(
        self,
        node_impl,
        node_path="bitcoind",
        rpcport=18443,
        network="regtest",
        rpcuser="bitcoin",
        rpcpassword="secret",
    ):
        try:
            super().__init__(
                node_impl,
                rpcport=rpcport,
                network=network,
                rpcuser=rpcuser,
                rpcpassword=rpcpassword,
            )
            self.node_path = node_path
            self.rpcconn.ipaddress = "localhost"
            self.status = "Down"
        except Exception as e:
            logger.exception(f"Failed to instantiate NodePlainController. Error: {e}")
            raise e

    def _start_node(
        self,
        cleanup_at_exit=True,
        cleanup_hard=False,
        datadir=None,
        log_stdout=False,
        extra_args=[],
    ):
        if datadir == None:
            datadir = tempfile.mkdtemp(
                prefix=f"specter_{self.node_impl}_regtest_plain_datadir_"
            )
        self.datadir = datadir

        node_cmd = self.construct_node_cmd(
            self.rpcconn,
            run_docker=False,
            datadir=datadir,
            log_stdout=log_stdout,
            node_path=self.node_path,
            network=self.network,
            extra_args=extra_args,
        )
        logger.debug("About to execute: {}".format(node_cmd))
        # exec will prevent creating a child-process and will make node_proc.terminate() work as expected
        self.node_proc = subprocess.Popen(
            ("exec " if platform.system() != "Windows" else "") + node_cmd,
            shell=True,
        )
        time.sleep(0.2)  # sleep 200ms (catch stdout of stupid errors)
        if not self.node_proc.poll() is None:
            # Might itself raise a SpecterError:
            debug_logs = self.get_debug_log()
            raise SpecterError(f"Could not start node due to:" + debug_logs)
        logger.debug(
            f"Running {self.node_impl}d-process with pid {self.node_proc.pid} in datadir {datadir}"
        )

        # This function is redirecting to the class.member as it needs a fixed parameterlist: (signal_number, stack)
        def cleanup_node_callback(signal_number=None, stack=None):
            self.cleanup_node(cleanup_hard, datadir)

        # If the node is shutdown via self.stop_node() (e.g. pytests) we need to know how hard we should do that
        self.cleanup_hard = cleanup_hard

        if cleanup_at_exit:
            logger.info(
                "Register function cleanup_node for atexit, SIGINT, and SIGTERM"
            )
            atexit.register(cleanup_node_callback)
            # This is for CTRL-C --> SIGINT
            signal.signal(signal.SIGINT, cleanup_node_callback)
            # This is for kill $pid --> SIGTERM
            signal.signal(signal.SIGTERM, cleanup_node_callback)

    def get_debug_log(self):

        logfile_location = os.path.join(
            self.datadir,
            self.network if self.network != "testnet" else "testnet3",
            "debug.log",
        )
        try:
            return "".join(get_last_lines_from_file(logfile_location))
        except FileNotFoundError as e:
            raise SpecterError(
                f"Could not find debug.log at {logfile_location}. Is that directory even existing?"
            )
        except Exception as e:
            logger.exception(f"Failed to get debug logs. Error: {e}")
            return "[Failed to get debug logs. Check the logs for details]"

    def cleanup_node(self, cleanup_hard=None, datadir=None):
        """KILLS or TERMINATES the node-process depending on cleanup_hard
        removes the datadir in case of KILL
        """
        returnvalue = True  # assume only the best
        if not hasattr(self, "datadir"):
            self.datadir = None

        if cleanup_hard == None:
            cleanup_hard = self.cleanup_hard
        if not hasattr(self, "node_proc"):
            logger.info("node process was not running")
            if cleanup_hard:
                logger.info(f"Removing node datadir: {datadir}")
                if datadir is None:
                    datadir = self.datadir
                shutil.rmtree(datadir, ignore_errors=True)
            returnvalue = False
        timeout = 50  # in secs
        logger.info(
            f"Cleaning up (cleanup_hard:{cleanup_hard} , datadir:{self.datadir})"
        )
        if cleanup_hard:
            try:
                self.node_proc.kill()  # might be usefull for e.g. testing. We can't wait for so long
                logger.info(
                    f"Killed {self.node_impl}d with pid {self.node_proc.pid}, Removing {self.datadir}"
                )
                shutil.rmtree(self.datadir, ignore_errors=True)
            except Exception as e:
                logger.error(e)
                returnvalue = False
        else:
            try:
                if not hasattr(self, "node_proc"):
                    returnvalue = False
                else:
                    self.node_proc.terminate()  # might take a bit longer than kill but it'll preserve block-height
                    logger.info(
                        f"Terminated {self.node_impl}d with pid {self.node_proc.pid}, waiting for termination (timeout {timeout} secs)..."
                    )
                    # self.node_proc.wait() # doesn't have a timeout
                    procs = psutil.Process().children()
                    for p in procs:
                        p.terminate()
                    _, alive = psutil.wait_procs(procs, timeout=timeout)
                    for p in alive:
                        logger.info(
                            f"{self.node_impl} did not terminated in time, killing!"
                        )
                        p.kill()
            except ProcessLookupError:
                # Bitcoind probably never came up or crashed.
                returnvalue = False
        if platform.system() == "Windows":
            subprocess.run("Taskkill /IM bitcoind.exe /F")
        return returnvalue

    def stop_node(self):
        success = self.cleanup_node()
        self.status = "Down"
        return success

    def check_existing(self):
        """other then in docker, we won't check on the "instance-level". This will return true if a
        node is running on the default port.
        """
        if not self.check_node(self.rpcconn):
            if self.status == "Running":
                self.status = "Down"
            return None
        else:
            self.status = "Running"
            return True


def find_node_executable(node_impl):
    '''Returns the path to a node_impl executable whereas node_impl is either "bitcoin" or "elements"'''
    if os.path.isfile(f"tests/{node_impl}/src/{node_impl}d"):
        # copied from conftest.py
        # always prefer the self-compiled bitcoind if existing
        return f"tests/{node_impl}/src/{node_impl}d"
    elif os.path.isfile(f"./tests/{node_impl}/bin/{node_impl}d"):
        # next take the self-installed binary if existing
        return f"tests/{node_impl}/bin/{node_impl}d"
    else:
        # First list files in the folders above:
        logger.warning(f"Couldn't find reasonable executable for {node_impl}")
        # hmmm, maybe we have a bitcoind on the PATH
        return which(f"{node_impl}d")


def fetch_wallet_addresses_for_mining(node_impl, data_folder):
    """parses all the wallet-jsons in the folder (default ~/.specter/wallets/regtest)
    and returns an array with the addresses
    """
    wallet_folder = (
        f"{data_folder}/wallets/{'regtest' if node_impl == 'bitcoin' else 'elreg'}"
    )
    wallets = load_jsons(wallet_folder)
    address_array = [value["address"] for key, value in wallets.items()]
    # remove duplicates
    address_array = list(dict.fromkeys(address_array))
    return address_array



import copy, random, json, re

from flask import (
    Flask,
    Blueprint,
    render_template,
    request,
    redirect,
    url_for,
    jsonify,
    flash,
)
from flask import current_app as app
from flask_babel import lazy_gettext as _
from flask_login import login_required, current_user
from mnemonic import Mnemonic
from ..devices.bitcoin_core import BitcoinCore
from ..helpers import is_testnet, generate_mnemonic, validate_mnemonic
from ..key import Key
from ..managers.device_manager import get_device_class
from ..managers.wallet_manager import purposes
from ..specter_error import handle_exception

rand = random.randint(0, 1e32)  # to force style refresh

# Setup endpoint blueprint
devices_endpoint = Blueprint("devices_endpoint", __name__)


################## New device #######################
# New device type
@devices_endpoint.route("/new_device_type/", methods=["GET", "POST"])
@login_required
def new_device_type():
    return render_template(
        "device/new_device/new_device_type.jinja",
        specter=app.specter,
        rand=rand,
    )


@devices_endpoint.route("/new_device_keys/<device_type>/", methods=["GET", "POST"])
@login_required
def new_device_keys(device_type):
    err = None
    mnemonic = ""
    passphrase = ""
    file_password = ""
    range_start = 0
    range_end = 1000
    existing_device = None
    if request.method == "POST":
        mnemonic = request.form.get("mnemonic", "")
        passphrase = request.form.get("passphrase", "")
        file_password = request.form.get("file_password", "")
        range_start = int(request.form.get("range_start", "0"))
        range_end = int(request.form.get("range_end", "1000"))
        existing_device = request.form.get("existing_device", None)
        if existing_device:
            device = app.specter.device_manager.get_by_alias(existing_device)
        else:
            device_name = request.form.get("device_name", "")
            if not device_name:
                err = _("Device name cannot be empty")
            elif device_name in app.specter.device_manager.devices_names:
                err = _("Device with this name already exists")
        xpubs_rows_count = int(request.form["xpubs_rows_count"]) + 1
        keys = []
        paths = []
        keys_purposes = []
        for i in range(0, xpubs_rows_count):
            purpose = request.form.get("xpubs-table-row-{}-purpose".format(i), "Custom")
            xpub = request.form.get("xpubs-table-row-{}-xpub-hidden".format(i), "-")
            path = request.form.get(
                "xpubs-table-row-{}-derivation-hidden".format(i), ""
            )
            if path != "":
                paths.append(path)
                keys_purposes.append(purpose)
            if xpub != "-":
                try:
                    keys.append(Key.parse_xpub(xpub, purpose=purpose))
                except:
                    err = _("Failed to parse these xpubs") + ":\n" + "\n".join(xpub)
                    break
        if not keys and not err:
            if device_type in ["bitcoincore", "elementscore"]:
                if not paths:
                    err = _("No paths were specified, please provide at least one.")
                if err is None:
                    if existing_device:
                        device.add_hot_wallet_keys(
                            mnemonic,
                            passphrase,
                            paths,
                            file_password,
                            app.specter.wallet_manager,
                            is_testnet(app.specter.chain),
                            keys_range=[range_start, range_end],
                            keys_purposes=keys_purposes,
                        )
                        flash(_("{} keys were added successfully").format(len(paths)))
                        return redirect(
                            url_for(
                                "devices_endpoint.device", device_alias=device.alias
                            )
                        )
                    device = app.specter.device_manager.add_device(
                        name=device_name, device_type=device_type, keys=[]
                    )
                    try:
                        device.setup_device(file_password, app.specter.wallet_manager)
                        device.add_hot_wallet_keys(
                            mnemonic,
                            passphrase,
                            paths,
                            file_password,
                            app.specter.wallet_manager,
                            is_testnet(app.specter.chain),
                            keys_range=[range_start, range_end],
                            keys_purposes=keys_purposes,
                        )
                        flash(_("{} was added successfully!").format(device_name))
                        return redirect(
                            url_for(
                                "devices_endpoint.device", device_alias=device.alias
                            )
                            + "?newdevice=true"
                        )
                    except Exception as e:
                        handle_exception(e)
                        flash(
                            _("Failed to setup hot wallet. Error: {}").format(e),
                            "error",
                        )
                        app.specter.device_manager.remove_device(
                            device,
                            app.specter.wallet_manager,
                            bitcoin_datadir=app.specter.bitcoin_datadir,
                            chain=app.specter.chain,
                        )
            else:
                err = _("xpubs list must not be empty")
        elif not err:
            if existing_device:
                device.add_keys(keys)
                flash(_("{} keys were added successfully").format(len(keys)))
                return redirect(
                    url_for("devices_endpoint.device", device_alias=device.alias)
                )
            device = app.specter.device_manager.add_device(
                name=device_name, device_type=device_type, keys=keys
            )
            if app.specter.is_liquid:
                return render_template(
                    "device/device_blinding_key.jinja",
                    new_device=True,
                    device=device,
                    error=err,
                    specter=app.specter,
                    rand=rand,
                )
            else:
                flash(_("{} was added successfully!").format(device_name))
                return redirect(
                    url_for("devices_endpoint.device", device_alias=device.alias)
                    + "?newdevice=true"
                )

    return render_template(
        "device/new_device/new_device_keys.jinja",
        device_class=get_device_class(device_type),
        mnemonic=mnemonic,
        passphrase=passphrase,
        file_password=file_password,
        range_start=range_start,
        range_end=range_end,
        existing_device=app.specter.device_manager.get_by_alias(existing_device)
        if existing_device
        else None,
        error=err,
        specter=app.specter,
        rand=rand,
    )


@devices_endpoint.route("/new_device_mnemonic/<device_type>/", methods=["GET", "POST"])
@login_required
def new_device_mnemonic(device_type):
    err = None
    strength = 128
    mnemonic = generate_mnemonic(
        strength=strength, language_code=app.get_language_code()
    )
    existing_device = None
    if request.method == "POST":
        if len(request.form["mnemonic"].split(" ")) not in [12, 15, 18, 21, 24]:
            err = _(
                "Invalid mnemonic entered: Must contain either: 12, 15, 18, 21, or 24 words."
            )
        if not validate_mnemonic(words=request.form["mnemonic"]):
            err = _("Invalid mnemonic entered.")
        range_start = int(request.form["range_start"])
        range_end = int(request.form["range_end"])
        if range_start > range_end:
            err = _("Invalid address range selected.")
        mnemonic = request.form["mnemonic"]
        passphrase = request.form["passphrase"]
        file_password = request.form["file_password"]
        existing_device = request.form.get("existing_device", None)
        if existing_device:
            existing_device = app.specter.device_manager.get_by_alias(existing_device)
        if not err:
            return render_template(
                "device/new_device/new_device_keys.jinja",
                device_class=get_device_class(device_type),
                mnemonic=mnemonic,
                passphrase=passphrase,
                file_password=file_password,
                range_start=range_start,
                range_end=range_end,
                existing_device=existing_device,
                error=err,
                specter=app.specter,
                rand=rand,
            )

    return render_template(
        "device/new_device/new_device_mnemonic.jinja",
        device_type=device_type,
        strength=strength,
        mnemonic=mnemonic,
        existing_device=existing_device,
        error=err,
        specter=app.specter,
        rand=rand,
    )


@devices_endpoint.route("/device_blinding_key/<device_alias>/", methods=["GET", "POST"])
@login_required
def device_blinding_key(device_alias):
    err = None
    try:
        device = app.specter.device_manager.get_by_alias(device_alias)
    except:
        return render_template(
            "base.jinja", error=_("Device not found"), specter=app.specter, rand=rand
        )
    if not device:
        return redirect(url_for("index"))
    if request.method == "POST":
        new_device = request.form.get("new_device", False)
        blinding_key = request.form.get("blinding_key")
        try:
            device.set_blinding_key(blinding_key)
            if not new_device:
                flash(_("Master blinding key was added successfully"))
            return redirect(
                url_for("devices_endpoint.device", device_alias=device.alias)
                + ("?newdevice=true" if new_device else "")
            )
        except Exception as e:
            flash(
                _("Invalid master blinding key! It should be in WIF or hex format."),
                "error",
            )

    return render_template(
        "device/device_blinding_key.jinja",
        device=device,
        error=err,
        specter=app.specter,
        rand=rand,
    )


# New device "manual" (deprecated)
@devices_endpoint.route("/new_device_manual/", methods=["GET", "POST"])
@login_required
def new_device_manual():
    err = None
    device_type = ""
    device_name = ""
    xpubs = ""
    strength = 128
    mnemonic = generate_mnemonic(
        strength=strength, language_code=app.get_language_code()
    )
    if request.method == "POST":
        action = request.form["action"]
        device_type = request.form["device_type"]
        device_name = request.form["device_name"]
        if action == "newcolddevice":
            if not device_name:
                err = _("Device name cannot be empty")
            elif device_name in app.specter.device_manager.devices_names:
                err = _("Device with this name already exists")
            xpubs = request.form["xpubs"]
            if not xpubs:
                err = _("xpubs name cannot be empty")
            keys, failed = Key.parse_xpubs(xpubs)
            if len(failed) > 0:
                err = _("Failed to parse these xpubs") + ":\n" + "\n".join(failed)
            if err is None:
                device = app.specter.device_manager.add_device(
                    name=device_name, device_type=device_type, keys=keys
                )
                return redirect(
                    url_for("devices_endpoint.device", device_alias=device.alias)
                )
        elif action == "newhotdevice":
            if not device_name:
                err = _("Device name cannot be empty")
            elif device_name in app.specter.device_manager.devices_names:
                err = _("Device with this name already exists")
            if len(request.form["mnemonic"].split(" ")) not in [12, 15, 18, 21, 24]:
                err = _(
                    "Invalid mnemonic entered: Must contain either: 12, 15, 18, 21, or 24 words."
                )

            if not validate_mnemonic(words=request.form["mnemonic"]):
                err = _("Invalid mnemonic entered.")
            range_start = int(request.form["range_start"])
            range_end = int(request.form["range_end"])
            if range_start > range_end:
                err = _("Invalid address range selected.")
            if err is None:
                mnemonic = request.form["mnemonic"]
                paths = [
                    l.strip()
                    for l in request.form["derivation_paths"].split("\n")
                    if len(l) > 0
                ]
                passphrase = request.form["passphrase"]
                file_password = request.form["file_password"]
                device = app.specter.device_manager.add_device(
                    name=device_name, device_type=device_type, keys=[]
                )
                try:
                    device.setup_device(file_password, app.specter.wallet_manager)
                    device.add_hot_wallet_keys(
                        mnemonic,
                        passphrase,
                        paths,
                        file_password,
                        app.specter.wallet_manager,
                        is_testnet(app.specter.chain),
                        keys_range=[range_start, range_end],
                    )
                    return redirect(
                        url_for("devices_endpoint.device", device_alias=device.alias)
                    )
                except Exception as e:
                    handle_exception(e)
                    flash(_("Failed to setup hot wallet. Error: {}").format(e), "error")
                    app.specter.device_manager.remove_device(
                        device,
                        app.specter.wallet_manager,
                        bitcoin_datadir=app.specter.bitcoin_datadir,
                        chain=app.specter.chain,
                    )
        elif action == "generatemnemonic":
            strength = int(request.form["strength"])
            mnemonic = generate_mnemonic(
                strength=strength, language_code=app.get_language_code()
            )
    return render_template(
        "device/new_device_manual.jinja",
        device_type=device_type,
        device_name=device_name,
        xpubs=xpubs,
        mnemonic=mnemonic,
        strength=strength,
        error=err,
        specter=app.specter,
        rand=rand,
    )


################## Device page #######################


@devices_endpoint.route("device/<device_alias>/", methods=["GET", "POST"])
@login_required
def device(device_alias):
    err = None
    try:
        device = app.specter.device_manager.get_by_alias(device_alias)
    except:
        return render_template(
            "base.jinja", error=_("Device not found"), specter=app.specter, rand=rand
        )
    if not device:
        return redirect(url_for("index"))
    wallets = device.wallets(app.specter.wallet_manager)
    if request.method == "POST":
        action = request.form["action"]
        if action == "forget":
            if len(wallets) != 0:
                # TODO: Long message strings like this should be moved into a template.
                err = (
                    _(
                        "Device could not be removed since it is used in wallets: {}"
                    ).format([wallet.name for wallet in wallets])
                    + "<br>"
                    + _(
                        "You must delete those wallets before you can remove this device."
                    )
                    + "<br>"
                    + _("You can delete a wallet from its Settings -> Advanced page.")
                )
            else:
                app.specter.device_manager.remove_device(
                    device,
                    app.specter.wallet_manager,
                    bitcoin_datadir=app.specter.bitcoin_datadir,
                    chain=app.specter.chain,
                )
                return redirect("")
        elif action == "delete_key":
            key = Key.from_json({"original": request.form["key"]})
            wallets_with_key = [w for w in wallets if key in w.keys]
            if len(wallets_with_key) != 0:
                # TODO: Long message strings like this should be moved into a template.
                err = (
                    _(
                        "Key could not be removed since it is used in wallets: {}"
                    ).format(", ".join([wallet.name for wallet in wallets_with_key]))
                    + "<br>"
                    + _("You must delete those wallets before you can remove this key.")
                    + "<br>"
                    + _("You can delete a wallet from its Settings -> Advanced page.")
                )
            else:
                device.remove_key(key)
        elif action == "rename":
            device_name = request.form["newtitle"]
            if not device_name:
                flash(_("Device name cannot be empty"), "error")
            elif device_name == device.name:
                pass
            elif device_name in app.specter.device_manager.devices_names:
                flash(_("Device already exists"), "error")
            else:
                device.rename(device_name)
        elif action == "add_keys":
            strength = 128
            mnemonic = generate_mnemonic(
                strength=strength, language_code=app.get_language_code()
            )
            if device.hot_wallet:
                return render_template(
                    "device/new_device/new_device_mnemonic.jinja",
                    mnemonic=mnemonic,
                    strength=strength,
                    existing_device=device,
                    device_alias=device_alias,
                    device_class=get_device_class(device.device_type),
                    device_type=device.device_type,
                    specter=app.specter,
                    rand=rand,
                )
            else:
                return render_template(
                    "device/new_device/new_device_keys.jinja",
                    existing_device=device,
                    device_alias=device_alias,
                    device_class=get_device_class(device.device_type),
                    specter=app.specter,
                    rand=rand,
                )
        elif action == "settype":
            device_type = request.form["device_type"]
            device.set_type(device_type)
    device = copy.deepcopy(device)

    def sort_accounts(k):
        # Ordering: 1) chain 2) account 3) purpose
        pattern = r"^m\/([0-9]+)h\/([0-9])h\/([0-9]+)h"
        if not k.derivation:
            return 0
        match = re.search(pattern, k.derivation)
        if not match:
            return 0
        return (
            int(match.group(1))
            + (int(match.group(2)) + 1) * 1000000
            + (int(match.group(3)) + 1) * 2000
        )

    device.keys.sort(key=sort_accounts, reverse=False)
    return render_template(
        "device/device.jinja",
        device=device,
        device_alias=device_alias,
        purposes=purposes,
        wallets=wallets,
        error=err,
        specter=app.specter,
        rand=rand,
    )



import logging

from cryptoadvance.specter.api.rest.base import (
    AdminResource,
    rest_resource,
)
from flask import current_app as app
from datetime import datetime

logger = logging.getLogger(__name__)


@rest_resource
class ResourceSpecter(AdminResource):
    """/api/v1alpha/specter"""

    endpoints = ["/v1alpha/specter"]

    def get(self):
        specter_data = app.specter

        return_dict = {
            "data_folder": specter_data.data_folder,
            "config": specter_data.config,
            "info": specter_data.info,
            "network_info": specter_data.network_info,
            "device_manager_datafolder": specter_data.device_manager.data_folder,
            "devices_names": specter_data.device_manager.devices_names,
            "wallets_names": specter_data.wallet_manager.wallets_names,
            "last_update": datetime.now().strftime("%m/%d/%Y, %H:%M:%S"),
        }

        # Include alias list for easy lookup between names and alias
        # Maybe there's a better way to create this
        wallets_alias = []
        alias_name = {}
        name_alias = {}
        for wallet in return_dict["wallets_names"]:
            alias = specter_data.wallet_manager.wallets[wallet].alias
            wallets_alias.append(alias)
            alias_name[alias] = wallet
            name_alias[wallet] = alias
        return_dict["alias_name"] = alias_name
        return_dict["name_alias"] = name_alias
        return_dict["wallets_alias"] = wallets_alias
        return return_dict



import json
import logging
import requests
from cryptoadvance.specter.managers.wallet_manager import WalletManager

from cryptoadvance.specter.specter_error import SpecterError

from ..helpers import is_testnet
from .descriptor import AddChecksum, Descriptor

logger = logging.getLogger(__name__)


class WalletImporter:
    """A class to create Wallets easily by json"""

    def __init__(self, wallet_json, specter, device_manager=None):
        """this will analyze the wallet_json and specifies self. ...:
        * wallet_name
        * recv_descriptor
        * cosigners_types
        from recv_descriptor and specter.chain:
        * descriptor
        from descriptor:
        * keys
        * cosigners
        * unknown_cosigners
        * unknown_cosigners_types
        """
        if device_manager is None:
            device_manager = specter.device_manager
        try:
            self.wallet_data = json.loads(wallet_json)
            (
                self.wallet_name,
                self.recv_descriptor,
                self.cosigners_types,
            ) = WalletImporter.parse_wallet_data_import(self.wallet_data)
        except Exception as e:
            logger.warning(f"Trying to import: {wallet_json}")
            raise SpecterError(f"Unsupported wallet import format:{e}")
        try:
            self.descriptor = Descriptor.parse(
                AddChecksum(self.recv_descriptor.split("#")[0]),
                testnet=is_testnet(specter.chain),
            )
            if self.descriptor is None:
                raise SpecterError(f"Invalid wallet descriptor. (returns None)")
        except Exception as e:
            raise SpecterError(f"Invalid wallet descriptor: {e}")
        if self.wallet_name in specter.wallet_manager.wallets_names:
            raise SpecterError(f"Wallet with the same name already exists")
        (
            self.keys,
            self.cosigners,
            self.unknown_cosigners,
            self.unknown_cosigners_types,
        ) = self.descriptor.parse_signers(device_manager.devices, self.cosigners_types)
        self.wallet_type = "multisig" if self.descriptor.multisig_N > 1 else "simple"

    def create_nonexisting_signers(self, device_manager, request_form):
        """creates non existinging signer via the device_manager
        request_form is a dict which needs (where X is an integer):
        * unknown_cosigner_X_name
        * unknown_cosigner_X_type
        Apart from adding the devices, it'll also modify:
        * self.keys and
        * self.cosigners
        """
        for i, (unknown_cosigner_key, label) in enumerate(self.unknown_cosigners):
            unknown_cosigner_name = request_form["unknown_cosigner_{}_name".format(i)]
            unknown_cosigner_type = request_form.get(
                "unknown_cosigner_{}_type".format(i), "other"
            )

            device = device_manager.add_device(
                name=unknown_cosigner_name,
                device_type=unknown_cosigner_type,
                keys=[unknown_cosigner_key],
            )
            logger.info(f"Creating device {device}")
            self.keys.append(unknown_cosigner_key)
            self.cosigners.append(device)

    def create_wallet(self, wallet_manager):
        """creates the wallet. Assumes all devices are there (create with create_nonexisting_signers)
        will also keypoolrefill and import_labels
        """
        try:
            self.wallet = wallet_manager.create_wallet(
                name=self.wallet_name,
                sigs_required=self.descriptor.multisig_M,
                key_type=self.descriptor.address_type,
                keys=self.keys,
                devices=self.cosigners,
            )
        except Exception as e:
            raise SpecterError(f"Failed to create wallet: {e}")
        logger.info(f"Created Wallet {self.wallet}")
        self.wallet.keypoolrefill(0, self.wallet.IMPORT_KEYPOOL, change=False)
        self.wallet.keypoolrefill(0, self.wallet.IMPORT_KEYPOOL, change=True)
        self.wallet.import_labels(self.wallet_data.get("labels", {}))
        return self.wallet

    def rescan_as_needed(self, specter):
        """will rescan the created wallet"""
        if not hasattr(self, "wallet"):
            raise Exception("called rescan_as_needed before create_wallet")
        potential_errors = None
        try:
            # get min of the two
            # if the node is still syncing
            # and the first block with tx is not there yet
            startblock = min(
                self.wallet_data.get("blockheight", specter.info.get("blocks", 0)),
                specter.info.get("blocks", 0),
            )
            # check if pruned
            if specter.info.get("pruned", False):
                newstartblock = max(startblock, specter.info.get("pruneheight", 0))
                if newstartblock > startblock:
                    potential_errors = SpecterError(
                        f"Using pruned node - we will only rescan from block {newstartblock}"
                    )
                    startblock = newstartblock
            self.wallet.rpc.rescanblockchain(startblock, timeout=1)
            logger.info("Rescanning Blockchain ...")
        except requests.exceptions.ReadTimeout:
            # this is normal behavior in our usecase
            pass
        except Exception as e:
            logger.error("Exception while rescanning blockchain: %r" % e)
            if potential_errors:
                potential_errors = SpecterError(
                    potential_errors
                    + " and "
                    + "Failed to perform rescan for wallet: %r" % e
                )
        self.wallet.getdata()
        if potential_errors:
            raise potential_errors

    @property
    def wallet_json(self):
        """self.wallet_data in json format"""
        return json.dumps(self.wallet_data)

    @classmethod
    def parse_wallet_data_import(cls, wallet_data):
        """Parses wallet JSON for import, takes JSON in a supported format
        and returns a tuple of wallet name, wallet descriptor, and cosigners types (electrum
        and newer Specter backups).
        Supported formats: Specter, Electrum, Account Map (Fully Noded, Gordian, Sparrow etc.)
        """
        cosigners_types = []

        # Specter-DIY format
        if "recv_descriptor" in wallet_data:
            wallet_name = wallet_data.get("name", "Imported Wallet")
            recv_descriptor = wallet_data.get("recv_descriptor", None)

        # Electrum multisig
        elif "x1/" in wallet_data:
            i = 1
            xpubs = ""
            while "x{}/".format(i) in wallet_data:
                d = wallet_data["x{}/".format(i)]
                xpubs += "[{}]{}/0/*,".format(
                    d["derivation"].replace("m", d["root_fingerprint"]), d["xpub"]
                )
                cosigners_types.append({"type": d["hw_type"], "label": d["label"]})
                i += 1
            xpubs = xpubs.rstrip(",")

            if "xpub" in wallet_data["x1/"]:
                wallet_type = WalletImporter.wallet_type_by_slip132_xpub(
                    wallet_data["x1/"]["xpub"]
                )
            else:
                raise Exception('"xpub" not found in "x1/" in Electrum backup json')

            required_sigs = int(wallet_data.get("wallet_type").split("of")[0])
            recv_descriptor = "{}(sortedmulti({}, {}))".format(
                wallet_type, required_sigs, xpubs
            )
            wallet_name = "Electrum {} of {}".format(required_sigs, i - 1)

        # Electrum singlesig
        # Here's a trimmed version of what an example Electrum file looks like (varies for different wallet types)
        # {
        #     "addr_history": {
        #         "tb1q2zqu0rxg70scs9umwwkjxvq7zdcle2ny6pda8v": [],
        #         "tb1q3p86mshyyrmzp2mw50eqeefxtkz8qfl8tsfgrs": [],
        #         "tb1q46u2rm20rp9rzauxxhdefregggag8cpelh0sqa": [
        #             [
        #                 "14f6bd3dec07c15773b29b4c7f7d3c913197840dfdf567b6b078e224f021b4ac",
        #                 2035683
        #             ],
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ]
        #         ],
        #         "tb1q4k59fgwhwey8k4rs2l80p73dwtza0z0gvpjjfc": [],
        #         "tb1q56f5wgc7wqeqygpfgua64gucafp3wp6293ggcg": [],
        #         "tb1q8uysvaz8dysekwws868knvxngh5qfjzrlxqf2j": [
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ]
        #         ],
        #         "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy": [
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qazyz882jeruqlppfralemm3r07pw6cpk426fhw": [],
        #         "tb1qch2j8ks2qt6thtf7wdpk6v9gquu8tfca6lsh0n": [],
        #         "tb1qcx2dultr5x9dc6jz0ye3wq240xfsdxnnl7vpsm": [
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qe686y2y39ncnafpzxxuyfztsy70thl9k4l4crk": [],
        #         "tb1qg33wumlly0mn33s776n8njqek3l2rmy8d97r02": [],
        #         "tb1qh48nqrg595g0vfqcw46kt8aun63t9puap5uww5": [],
        #         "tb1qhlgkkqtdglq64xa8wnwz4446mzj97g0jq9ck65": [],
        #         "tb1qjrd86fxf5chezp2pkjzv8v45rzxe8japwxatt2": [],
        #         "tb1qlty269zlah8h9afx0cxvn05qrzyndvnz28ae6d": [],
        #         "tb1qmcnxezw6qh8vys5sxpwfcqgs2t3gyuy7setdky": [],
        #         "tb1qphhms4gdrfpxh5ed2zekhvnwrk84jk5z56qj82": [],
        #         "tb1qq5huaklp886yyqmxju47xm42dgn5demx5yd848": [],
        #         "tb1qqfp0y5qtf0cx7hzat7r3g69gxx96fkg3vnu4cz": [],
        #         "tb1qqu0p6m8ux64kkflx9wa5sc8k39v8crfsn96qky": [],
        #         "tb1qtu73qp3eneht3rfauwz3rttvs7m7l5lahkq66m": [],
        #         "tb1qtvca6a7gp76km7k5kyg7e6lxhvx3qme6x9p53a": [],
        #         "tb1qtxh9dqg07p32e93vpzhfmv7sgp9je5phps5gpy": [],
        #         "tb1qu0wcfv6n96xm6nx5lczcc4gc5h3fjmsfqsmaar": [],
        #         "tb1qx0ld4mml0hkrwuejmetl77rma08ehc4d75qruy": [],
        #         "tb1qxf7jgjd8k0nmm8g8vclx9zg5drr3hfls3p8c6m": [
        #             [
        #                 "23196c7657b60d1fde716c789881ffeee93554b9040c200589c65ceeac051cfb",
        #                 2035683
        #             ],
        #             [
        #                 "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde",
        #                 0
        #             ]
        #         ],
        #         "tb1qy0pklnhg7rks54gzlu003muyllym3xflkd8e7m": [],
        #         "tb1qyt9mkqgf5gfj0lxhmuzv4xamnra8jlxyhpr5x9": [],
        #         "tb1qz9fe873jvrwd0clcknjgj8z6t3uylc50lgvh2q": []
        #     },
        #     "addresses": {
        #         "change": [
        #             "tb1qxf7jgjd8k0nmm8g8vclx9zg5drr3hfls3p8c6m",
        #             "tb1qcx2dultr5x9dc6jz0ye3wq240xfsdxnnl7vpsm",
        #             "tb1q3p86mshyyrmzp2mw50eqeefxtkz8qfl8tsfgrs",
        #             "tb1qe686y2y39ncnafpzxxuyfztsy70thl9k4l4crk",
        #             "tb1qlty269zlah8h9afx0cxvn05qrzyndvnz28ae6d",
        #             "tb1q56f5wgc7wqeqygpfgua64gucafp3wp6293ggcg",
        #             "tb1qqfp0y5qtf0cx7hzat7r3g69gxx96fkg3vnu4cz",
        #             "tb1qhlgkkqtdglq64xa8wnwz4446mzj97g0jq9ck65",
        #             "tb1q4k59fgwhwey8k4rs2l80p73dwtza0z0gvpjjfc",
        #             "tb1qq5huaklp886yyqmxju47xm42dgn5demx5yd848"
        #         ],
        #         "receiving": [
        #             "tb1q46u2rm20rp9rzauxxhdefregggag8cpelh0sqa",
        #             "tb1q8uysvaz8dysekwws868knvxngh5qfjzrlxqf2j",
        #             "tb1qtvca6a7gp76km7k5kyg7e6lxhvx3qme6x9p53a",
        #             "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy",
        #             "tb1qtu73qp3eneht3rfauwz3rttvs7m7l5lahkq66m",
        #             "tb1qu0wcfv6n96xm6nx5lczcc4gc5h3fjmsfqsmaar",
        #             "tb1qx0ld4mml0hkrwuejmetl77rma08ehc4d75qruy",
        #             "tb1qazyz882jeruqlppfralemm3r07pw6cpk426fhw",
        #             "tb1qz9fe873jvrwd0clcknjgj8z6t3uylc50lgvh2q",
        #             "tb1qy0pklnhg7rks54gzlu003muyllym3xflkd8e7m",
        #             "tb1qh48nqrg595g0vfqcw46kt8aun63t9puap5uww5",
        #             "tb1qjrd86fxf5chezp2pkjzv8v45rzxe8japwxatt2",
        #             "tb1qg33wumlly0mn33s776n8njqek3l2rmy8d97r02",
        #             "tb1qmcnxezw6qh8vys5sxpwfcqgs2t3gyuy7setdky",
        #             "tb1q2zqu0rxg70scs9umwwkjxvq7zdcle2ny6pda8v",
        #             "tb1qyt9mkqgf5gfj0lxhmuzv4xamnra8jlxyhpr5x9",
        #             "tb1qphhms4gdrfpxh5ed2zekhvnwrk84jk5z56qj82",
        #             "tb1qqu0p6m8ux64kkflx9wa5sc8k39v8crfsn96qky",
        #             "tb1qch2j8ks2qt6thtf7wdpk6v9gquu8tfca6lsh0n",
        #             "tb1qtxh9dqg07p32e93vpzhfmv7sgp9je5phps5gpy"
        #         ]
        #     },
        #     "fiat_value": {},
        #     "frozen_coins": {},
        #     "imported_channel_backups": {},
        #     "keystore": {
        #         "derivation": "m/0'",
        #         "pw_hash_version": 1,
        #         "root_fingerprint": "8e0c95ad",
        #         "seed": "toss despair unaware moral shield process morning cook enlist park talent snack",
        #         "seed_type": "segwit",
        #         "type": "bip32",
        #         "xprv": "...",
        #         "xpub": "vpub5VGXXixD2pHLFtcKtCF57e8mx2JW6fie8VydXijC8sRKAL4RshgjEmzbmV915NeVB9pd23DVYem6zWM7HXFLNwaffNVHowdD9SJWwESyQhp"
        #     },
        #     "labels": {
        #         "14f6bd3dec07c15773b29b4c7f7d3c913197840dfdf567b6b078e224f021b4ac": "this is a txid label",
        #         "c9a54c381293a387a8c1c90e542d11cccf42cd0aeab20be3d1fcd10b465dcfde": "this is an address label",
        #         "tb1q8ydkp6qhq79ktwpcwjgt9dl7eer32vpw8gvduy": "this is an address label"
        #     },
        #     "lightning_payments": {},
        #     "lightning_preimages": {},
        #     "onchain_channel_backups": {},
        #     "qt-console-history": [],
        #     "seed_type": "segwit",
        #     "seed_version": 41,
        #     "stored_height": 2035683,
        #     "submarine_swaps": {},
        #     "use_encryption": false,
        #     "wallet_type": "standard",
        # }
        elif "keystore" in wallet_data:
            wallet_name = wallet_data["keystore"].get("label", "Electrum Wallet")

            if "xpub" in wallet_data["keystore"]:
                wallet_type = cls.wallet_type_by_slip132_xpub(
                    wallet_data["keystore"]["xpub"], is_multisig=False
                )
            else:
                raise Exception(
                    '"xpub" not found in "keystore" in Electrum backup json'
                )
            recv_descriptor = "{}({})".format(
                wallet_type,
                "[{}]{}/0/*,".format(
                    wallet_data["keystore"]["derivation"].replace(
                        "m", wallet_data["keystore"]["root_fingerprint"]
                    ),
                    wallet_data["keystore"]["xpub"],
                ),
            )
            cosigners_types = [
                {
                    "type": wallet_data["keystore"].get("hw_type", "other"),
                    "label": wallet_data["keystore"].get("label", "Electrum Wallet"),
                }
            ]

        # Current Specter backups
        else:
            # Newer exports are able to reinitialize device types but stay backwards
            #   compatible with older backups.
            if "devices" in wallet_data:
                cosigners_types = wallet_data["devices"]

            wallet_name = wallet_data.get("label", "Imported Wallet")
            recv_descriptor = wallet_data.get("descriptor", None)

        return (wallet_name, recv_descriptor, cosigners_types)

    @classmethod
    def wallet_type_by_slip132_xpub(cls, xpub, is_multisig=True):
        """
        see: https://github.com/satoshilabs/slips/blob/master/slip-0132.md
        Electrum backups use SLIP-132 but note that other wallets don't make the same
        guarantee.
        """
        if is_multisig:
            if xpub.startswith("xpub") or xpub.startswith("tpub"):
                return "sh"
            elif xpub.startswith("Ypub") or xpub.startswith("Upub"):
                return "sh-wsh"
            elif xpub.startswith("Zpub") or xpub.startswith("Vpub"):
                return "wsh"
        else:
            if xpub.startswith("xpub") or xpub.startswith("tpub"):
                return "pkh"
            elif xpub.startswith("ypub") or xpub.startswith("upub"):
                return "sh-wpkh"
            elif xpub.startswith("zpub") or xpub.startswith("vpub"):
                return "wpkh"
        raise Exception(f"Unhandled xpub type: {xpub}")



"""
Manages the list of transactions for the wallet
"""
import os
from .persistence import write_csv, read_csv
from .helpers import get_address_from_dict
from embit.transaction import Transaction
from embit.liquid.networks import get_network
import json
import logging
from .util.tx import decoderawtransaction

logger = logging.getLogger(__name__)


def parse_arr(v):
    if not isinstance(v, str):
        return v
    try:
        return json.loads(v.replace("'", '"'))
    except:
        return v


class TxItem(dict):
    TransactionCls = Transaction
    columns = [
        "txid",  # str, txid in hex
        "blockhash",  # str, blockhash, None if not confirmed
        "blockheight",  # int, blockheight, None if not confirmed
        "time",  # int (timestamp in seconds), time received
        "bip125-replaceable",  # str ("yes" / "no"), whatever RBF is enabled for the transaction
        "conflicts",  # rbf conflicts, list of txids
        "vsize",
        "category",
        "address",
        "amount",
        "ismine",
    ]
    type_converter = [
        str,
        str,
        int,
        int,
        str,
        parse_arr,
        int,
        str,
        parse_arr,
        parse_arr,
        bool,
    ]

    def __init__(self, rpc, addresses, rawdir, chain, **kwargs):
        self.rpc = rpc
        self._addresses = addresses
        self.rawdir = rawdir
        self.chain = chain
        # copy
        kwargs = dict(**kwargs)
        # replace with None or convert
        for i, k in enumerate(self.columns):
            v = kwargs.get(k, "")
            kwargs[k] = None if v in ["", None] else self.type_converter[i](v)

        super().__init__(**kwargs)
        self._tx = None
        # if we have hex data
        if kwargs.get("hex"):
            self._tx = self.TransactionCls.from_string(kwargs["hex"])

    @property
    def fname(self):
        return os.path.join(self.rawdir, self.txid + ".bin")

    @property
    def tx(self):
        if not self._tx:
            # Get transaction from file if we don't have it cached.
            # We cache transactions to self._tx
            # only when new tx is added before dump() is called
            try:
                if os.path.isfile(self.fname):
                    with open(self.fname, "rb") as f:
                        tx = self.TransactionCls.read_from(f)
                        self._tx = tx
                        return tx
            except Exception as e:
                logger.error(e)
        # if we failed to load tx from file - load it from RPC
        if not self._tx:
            # get transaction from rpc
            try:
                res = self.rpc.gettransaction(self.txid)
                tx = self.TransactionCls.from_string(res["hex"])
                self._tx = tx
                return tx
            except Exception as e:
                logger.error(e)
        return self._tx

    def dump(self):
        """Dumps transaction in binary to the folder if it's not there"""
        # nothing to do if file exists or we don't have binary tx
        if os.path.isfile(self.fname) or not self._tx:
            return
        # create dir if it doesn't exist
        if not os.path.isdir(self.rawdir):
            os.mkdir(self.rawdir)
        with open(self.fname, "wb") as f:
            self.tx.write_to(f)
        # clear cached tx as we saved the transaction to file
        self._tx = None

    @property
    def txid(self):
        return self["txid"]

    @property
    def hex(self):
        return str(self.tx)

    def __str__(self):
        return self.txid

    def __repr__(self):
        return f"TxItem({str(self)})"

    def __dict__(self):
        return {
            "txid": self["txid"],
            "blockhash": self["blockhash"],
            "blockheight": self["blockheight"],
            "time": self["time"],
            "conflicts": self["conflicts"],
            "bip125-replaceable": self["bip125-replaceable"],
            "vsize": self["vsize"],
            "category": self["category"],
            "address": self["address"],
            "amount": self["amount"],
            "ismine": self["ismine"],
        }


class TxList(dict):
    ItemCls = TxItem  # for inheritance

    def __init__(self, path, rpc, addresses, chain):
        self.chain = chain
        self.path = path
        # folder to store transactions in binary form
        self.rawdir = path.replace(".csv", "_raw")
        self.rpc = rpc
        self._addresses = addresses
        txs = []
        file_exists = False
        try:
            if os.path.isfile(self.path):
                txs = read_csv(
                    self.path,
                    self.ItemCls,
                    self.rpc,
                    self._addresses,
                    self.rawdir,
                    self.chain,
                )
                for tx in txs:
                    self[tx.txid] = tx
                file_exists = True
        except Exception as e:
            logger.error(e)
        self._file_exists = file_exists

    def save(self):
        # check if we have at least one transaction
        if self:
            # Dump all transactions to binary files
            # This happens only if they have not been dumped before
            for tx in self.values():
                tx.dump()
            write_csv(self.path, list(self.values()), self.ItemCls)
        self._file_exists = True

    def gettransaction(self, txid, blockheight=None, decode=False, full=True):
        """
        Will ask Bitcoin Core for a transaction if blockheight is None or txid not known
        Provide blockheight or 0 if you don't care about confirmations number
        to avoid RPC calls.
        full=True will add a "hex" key
        decode=True will decode transaction similar to Core's decoderawtransaction
        """
        # if we don't know blockheigth or transaction
        # we get it from rpc
        if blockheight is None or txid not in self:
            tx = self.rpc.gettransaction(txid)
            if "time" not in tx:
                tx["time"] = tx["timereceived"]
            # save if we don't know about this tx
            if txid not in self:
                self.add({txid: tx})
        # lookup tx in cache, it should be there already
        tx = self[txid]
        res = dict(**tx)
        if full:
            res["hex"] = tx.hex
        if decode:
            res.update(self.decoderawtransaction(tx.hex))
        return res

    def decoderawtransaction(self, txhex):
        return decoderawtransaction(txhex, self.chain)

    def add(self, txs):
        """
        adds transactions to the list without any rpc calls.
        assuming tx is already watched by the wallet.
        Transactions should be a dict with dicts with fields:
        "<txid>": {
            "txid", - hex txid
            "hex",  - hex raw transaction
            "blockheight", - int blockheight if confirmed, None otherwise
            "blockhash", - str blockhash if confirmed, None otherwise
            "time", - int unix timestamp in seconds when tx was received
            "conflicts", - list of txids spending the same inputs (rbf)
            "bip125-replaceable", - str ("yes" or "no") - is rbf enabled for this tx
        }
        """
        # here we store all addresses in transactions
        # to set them used later
        addresses = []
        # first we add all transactions to cache
        for txid in txs:
            tx = txs[txid]
            # find minimal from 3 times:
            maxtime = 10445238000  # TODO: change after 31 dec 2300 lol
            time = min(
                tx.get("blocktime", maxtime),
                tx.get("timereceived", maxtime),
                tx.get("time", maxtime),
            )
            obj = {
                "txid": txid,
                "blockheight": tx.get("blockheight", None),
                "blockhash": tx.get("blockhash", None),
                "time": time,
                "conflicts": tx.get("walletconflicts", []),
                "bip125-replaceable": tx.get("bip125-replaceable", "no"),
                "hex": tx.get("hex", None),
            }
            txitem = self.ItemCls(
                self.rpc, self._addresses, self.rawdir, self.chain, **obj
            )
            self[txid] = txitem
            if txitem.tx:
                for vout in txitem.tx.vout:
                    try:
                        addr = vout.script_pubkey.address(get_network(self.chain))
                        if addr not in addresses:
                            addresses.append(addr)
                    except:
                        pass  # maybe not an address, but a raw script?
        self._addresses.set_used(addresses)
        # detect category, amounts and addresses
        for tx in [self[txid] for txid in self if txid in txs]:
            self.fill_missing(tx)
        self.save()

    def fill_missing(self, tx):
        raw_tx = self.decoderawtransaction(tx.hex)
        tx["vsize"] = raw_tx["vsize"]

        category = ""
        addresses = []
        amounts = {}
        inputs_mine_count = 0
        for vin in raw_tx["vin"]:
            # coinbase tx
            if (
                vin["txid"]
                == "0000000000000000000000000000000000000000000000000000000000000000"
            ):
                category = "generate"
                break
            if vin["txid"] in self:
                try:
                    address = get_address_from_dict(
                        self.decoderawtransaction(self[vin["txid"]].hex)["vout"][
                            vin["vout"]
                        ]
                    )
                    address_info = self._addresses.get(address, None)
                    if address_info and not address_info.is_external:
                        inputs_mine_count += 1
                except Exception as e:
                    logger.error(e)
                    continue

        outputs_mine_count = 0
        for out in raw_tx["vout"]:
            try:
                address = get_address_from_dict(out)
            except Exception as e:
                # couldn't get address...
                logger.error(e)
                continue
            address_info = self._addresses.get(address)
            if address_info and not address_info.is_external:
                outputs_mine_count += 1
            addresses.append(address)
            amounts[address] = out.get("value", 0)

        if inputs_mine_count:
            if outputs_mine_count == len(raw_tx["vout"]):
                category = "selftransfer"
                # remove change addresses from the dest list
                addresses2 = [
                    address
                    for address in addresses
                    if self._addresses.get(address, None)
                    and not self._addresses[address].change
                ]
                # use new list only if it's not empty
                if addresses2:
                    addresses = addresses2
            else:
                category = "send"
                addresses = [
                    address
                    for address in addresses
                    if not self._addresses.get(address, None)
                    or self._addresses[address].is_external
                ]
        else:
            if not category:
                category = "receive"
            addresses = [
                address
                for address in addresses
                if self._addresses.get(address, None)
                and not self._addresses[address].is_external
            ]

        amounts = [amounts[address] for address in addresses]

        if len(addresses) == 1:
            addresses = addresses[0]
            amounts = amounts[0]

        tx["category"] = category
        tx["address"] = addresses
        tx["amount"] = amounts
        if not addresses:
            tx["ismine"] = False
        else:
            tx["ismine"] = True

    def load(self, arr):
        """
        Load transactions to Core with merkle proofs to avoid rescan
        arr should be a dict with dicts:
        "<txid>": {
            "txid",         - hex txid
            "hex",          - optional, raw hex transactions
            "merkle_proof", - optional, hex merkle proof
            "blockheight",  - optional, block height
            "blockhash",    - optional, block hash
            "time",         - optional, time received
        }
        If optional fields are not found we will try to get them from rpc calls.
        If we fail - transaction will not be added.
        Retuns an dict of dicts:
        {
            <txid>: {"success": True}, on success,
            <txid>: {"success": False, "error": "reason"}, if failed to import transaction
        }
        """
        # TODO: how to handle unconfirmed?
        # try to get raw transactions and tx details if not present
        # self.save()
        return [{"success": False, "error": "not implemented"} for a in arr]

    @property
    def file_exists(self):
        return self._file_exists and os.path.isfile(self.path)



"""
Blockstream Jade Devices
************************
"""

from .jadepy.jade import JadeAPI, JadeError
from serial.tools import list_ports

from typing import List, Union
from hwilib.descriptor import PubkeyProvider
from hwilib.hwwclient import HardwareWalletClient
from hwilib.errors import (
    ActionCanceledError,
    BadArgumentError,
    DeviceConnectionError,
    DeviceFailureError,
    UnavailableActionError,
    common_err_msgs,
    handle_errors,
)
from hwilib.common import (
    AddressType,
    Chain,
)
from hwilib.key import ExtendedKey, parse_path
from hwilib.psbt import PSBT
from hwilib.tx import CTransaction
from hwilib._script import (
    is_p2sh,
    is_p2wpkh,
    is_p2wsh,
    is_witness,
)

import logging
import os

# embit-related things
from embit import ec, bip32
from embit.psbt import PSBT
from embit import script
from embit.liquid.pset import PSET
from embit import hashes
from embit.util import secp256k1
from embit.liquid.finalizer import finalize_psbt

JADE_VENDOR_ID = 0x10C4
JADE_DEVICE_ID = 0xEA60

py_enumerate = (
    enumerate  # To use the enumerate built-in, since the name is overridden below
)

logger = logging.getLogger(__name__)


def jade_exception(f):
    def func(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except ValueError as e:
            raise BadArgumentError(str(e))
        except JadeError as e:
            if e.code == -32000:  # CBOR_RPC_USER_CANCELLED
                raise ActionCanceledError("{} canceled by user".format(f.__name__))
            elif e.code == -32602:  # CBOR_RPC_BAD_PARAMETERS
                raise BadArgumentError(e.message)
            elif e.code == -32603:  # CBOR_RPC_INTERNAL_ERROR
                raise DeviceFailureError(e.message)
            elif e.code == -32002:  # CBOR_RPC_HW_LOCKED
                raise DeviceConnectionError("Device is locked")
            elif e.code == -32003:  # CBOR_RPC_NETWORK_MISMATCH
                raise DeviceConnectionError("Network/chain selection error")
            elif e.code in [
                -32600,
                -32601,
                -32001,
            ]:  # CBOR_RPC_INVALID_REQUEST, CBOR_RPC_UNKNOWN_METHOD, CBOR_RPC_PROTOCOL_ERROR
                raise DeviceConnectionError("Messaging/communiciation error")
            else:
                raise e

    return func


# This class extends the HardwareWalletClient for Blockstream Jade specific things
class JadeClient(HardwareWalletClient):

    NETWORKS = {Chain.MAIN: "mainnet", Chain.TEST: "testnet", Chain.REGTEST: "regtest"}
    liquid_network = None

    def set_liquid_network(self, chain):
        self.liquid_network = "liquid" if chain == "liquidv1" else "localtest-liquid"

    def _network(self):
        if self.liquid_network:
            return self.liquid_network
        return JadeClient.NETWORKS.get(self.chain, "mainnet")

    ADDRTYPES = {
        AddressType.LEGACY: "pkh(k)",
        AddressType.WIT: "wpkh(k)",
        AddressType.SH_WIT: "sh(wpkh(k))",
    }

    @staticmethod
    def _convertAddrType(addrType):
        return JadeClient.ADDRTYPES[addrType]

    def __init__(self, path: str, password: str = "", expert: bool = False) -> None:
        super(JadeClient, self).__init__(path, password, expert)
        self.jade = JadeAPI.create_serial(path)
        self.jade.connect()

        # Push some host entropy into jade
        self.jade.add_entropy(os.urandom(32))

        # Do the PIN thing if required
        # NOTE: uses standard 'requests' networking to connect to blind pinserver
        try:
            while not self.jade.auth_user(self._network()):
                logging.debug("Incorrect PIN provided")
        except:
            try:
                self.chain = Chain.TEST
                while not self.jade.auth_user(self._network()):
                    logging.debug("Incorrect PIN provided")
            except:
                self.chain = Chain.REGTEST
                while not self.jade.auth_user(self._network()):
                    logging.debug("Incorrect PIN provided")

    # Retrieves the public key at the specified BIP 32 derivation path
    @jade_exception
    def get_pubkey_at_path(self, bip32_path: str) -> ExtendedKey:
        path = parse_path(bip32_path)
        xpub = self.jade.get_xpub(self._network(), path)
        ext_key = ExtendedKey.deserialize(xpub)
        return ext_key

    @jade_exception
    def get_master_blinding_key(self) -> str:
        mbk = self.jade.get_master_blinding_key()
        assert len(mbk) == 32
        bkey = ec.PrivateKey(mbk)
        return bkey.wif()

    # Walk the PSBT looking for inputs we can sign.  Push any signatures into the
    # 'partial_sigs' map in the input, and return the updated PSBT.
    @jade_exception
    def sign_tx(self, tx: PSBT) -> PSBT:
        """
        Sign a transaction with the Blockstream Jade.

        - Jade can only be used to sign single-key inputs at this time.   It cannot sign multisig or arbitrary scripts.
        """

        c_txn = CTransaction(tx.tx)
        master_fp = self.get_master_fingerprint()
        signing_pubkeys = [None] * len(tx.inputs)

        # Signing input details
        jade_inputs = []
        for n_vin, (txin, psbtin) in py_enumerate(zip(c_txn.vin, tx.inputs)):
            # Get bip32 path to use to sign, if required for this input
            path = None
            for pubkey, origin in psbtin.hd_keypaths.items():
                if origin.fingerprint == master_fp and len(origin.path) > 0:
                    # Our input
                    if (
                        pubkey not in psbtin.partial_sigs
                        or not psbtin.partial_sigs[pubkey]
                    ):
                        # hw to sign this input - it is not already signed
                        signing_pubkeys[n_vin] = pubkey
                        path = origin.path

            # Get the tx and prevout/scriptcode
            utxo = None
            input_txn_bytes = None
            if psbtin.witness_utxo:
                utxo = psbtin.witness_utxo
            if psbtin.non_witness_utxo:
                if txin.prevout.hash != psbtin.non_witness_utxo.sha256:
                    raise BadArgumentError(
                        "Input {} has a non_witness_utxo with the wrong hash".format(
                            n_vin
                        )
                    )
                utxo = psbtin.non_witness_utxo.vout[txin.prevout.n]
                input_txn_bytes = psbtin.non_witness_utxo.serialize_without_witness()

            scriptcode = utxo.scriptPubKey

            if is_p2sh(scriptcode):
                scriptcode = psbtin.redeem_script

            witness_input, witness_version, witness_program = is_witness(scriptcode)

            if witness_input:
                if is_p2wsh(scriptcode):
                    scriptcode = psbtin.witness_script
                elif is_p2wpkh(scriptcode):
                    scriptcode = b"\x76\xa9\x14" + witness_program + b"\x88\xac"
                else:
                    scriptcode = None

            # Build the input and add to the list
            jade_inputs.append(
                {
                    "is_witness": witness_input,
                    "input_tx": input_txn_bytes,
                    "script": scriptcode,
                    "path": path,
                }
            )

        # Change output details
        # This is optional, in that if we send it Jade validates the change output script
        # and the user need not confirm that ouptut.  If not passed the change output must
        # be confirmed by the user on the hwwallet screen, like any other spend output.
        change = [None] * len(tx.outputs)
        for n_vout, (txout, psbtout) in py_enumerate(zip(c_txn.vout, tx.outputs)):
            for pubkey, origin in psbtout.hd_keypaths.items():
                # Considers 'our' outputs as change as far as Jade is concerned
                # ie. can be auto-confirmed.
                # Is this ok, or should check path also, assuming bip44-like ?
                if origin.fingerprint == master_fp and len(origin.path) > 0:
                    addr_type = None
                    if txout.is_p2pkh():
                        addr_type = AddressType.LEGACY
                    elif txout.is_witness()[0] and not txout.is_p2wsh():
                        addr_type = AddressType.WIT
                    elif txout.is_p2sh():
                        addr_type = AddressType.SH_WIT  # is it really though ?

                    if addr_type:
                        addr_type = self._convertAddrType(addr_type)
                        change[n_vout] = {"path": origin.path, "variant": addr_type}

        # The txn itself
        txn_bytes = c_txn.serialize_without_witness()

        # Request Jade generate the signatures for our inputs.
        # Change details are passed to be validated on the hw (user does not confirm)
        signatures = self.jade.sign_tx(self._network(), txn_bytes, jade_inputs, change)

        # Push sigs into PSBT structure as appropriate
        for psbtin, pubkey, sig in zip(tx.inputs, signing_pubkeys, signatures):
            if pubkey and sig:
                psbtin.partial_sigs[pubkey] = sig

        # Return the updated psbt
        return tx

    # Sign message, confirmed on device
    @jade_exception
    def sign_message(self, message: Union[str, bytes], bip32_path: str) -> str:
        path = parse_path(bip32_path)
        if isinstance(message, bytes) or isinstance(message, bytearray):
            message = message.decode("utf-8")
        signature = self.jade.sign_message(path, message)
        return signature

    # Display address of specified type on the device. Only supports single-key based addresses atm.
    @jade_exception
    def display_singlesig_address(self, bip32_path: str, addr_type: AddressType) -> str:
        path = parse_path(bip32_path)
        addr_type = self._convertAddrType(addr_type)
        address = self.jade.get_receive_address(
            self._network(), path, variant=addr_type
        )
        return address

    def display_multisig_address(
        self, threshold: int, pubkeys: List[PubkeyProvider], addr_type: AddressType
    ) -> str:
        """
        The Blockstream Jade does not support multisig addresses.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support generic multisig P2SH address display"
        )

    # Setup a new device
    def setup_device(self, label="", passphrase=""):
        """
        The Blockstream Jade does not support setup via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support software setup"
        )

    # Wipe this device
    def wipe_device(self):
        """
        The Blockstream Jade does not support wiping via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support wiping via software"
        )

    # Restore device from mnemonic or xprv
    def restore_device(self, label="", word_count=24):
        """
        The Blockstream Jade does not support restoring via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support restoring via software"
        )

    # Begin backup process
    def backup_device(self, label="", passphrase=""):
        """
        The Blockstream Jade does not support backing up via software.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support creating a backup via software"
        )

    # Close the device
    def close(self):
        self.jade.disconnect()

    # Prompt pin
    def prompt_pin(self):
        """
        The Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not need a PIN sent from the host"
        )

    # Send pin
    def send_pin(self, pin):
        """
        The Blockstream Jade does not need a PIN sent from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not need a PIN sent from the host"
        )

    # Toggle passphrase
    def toggle_passphrase(self):
        """
        The Blockstream Jade does not support toggling passphrase from the host.

        :raises UnavailableActionError: Always, this function is unavailable
        """
        raise UnavailableActionError(
            "The Blockstream Jade does not support toggling passphrase from the host"
        )

    def _blind(self, pset, seed: bytes = None):
        if seed is None:
            seed = pset.unknown.get(b"\xfc\x07specter\x00", os.urandom(32))
        txseed = pset.txseed(seed)
        # assign blinding factors to all outputs
        blinding_outs = []
        commitments = []
        # because we do sha once (cause taproot), and they want sha twice
        hash_prevouts = hashes.sha256(pset.blinded_tx.hash_prevouts())
        last_i = 0
        last_commitment = {}
        for i, out in py_enumerate(pset.outputs):
            # skip ones where we don't need blinding
            if out.blinding_pubkey is None:
                commitments.append(None)
                continue
            commitment = self.jade.get_commitments(
                bytes(reversed(out.asset)), out.value, hash_prevouts, i, vbf=None
            )
            commitment["blinding_key"] = out.blinding_pubkey
            commitments.append(commitment)
            last_i = i
            last_commitment = commitments[-1]
            out.asset_blinding_factor = commitment["abf"]
            out.value_blinding_factor = commitment["vbf"]
            blinding_outs.append(out)
        if len(blinding_outs) == 0:
            raise Exception("Nothing to blind")
        # calculate last vbf
        vals = [sc.value for sc in pset.inputs + blinding_outs]
        abfs = [
            sc.asset_blinding_factor or b"\x00" * 32
            for sc in pset.inputs + blinding_outs
        ]
        vbfs = [
            sc.value_blinding_factor or b"\x00" * 32
            for sc in pset.inputs + blinding_outs
        ]
        last_vbf = secp256k1.pedersen_blind_generator_blind_sum(
            vals, abfs, vbfs, len(pset.inputs)
        )
        last_out = blinding_outs[-1]
        new_last_commitment = self.jade.get_commitments(
            bytes(reversed(last_out.asset)),
            last_out.value,
            hash_prevouts,
            last_i,
            vbf=last_vbf,
        )
        # check abf didn't change
        assert new_last_commitment["abf"] == last_out.asset_blinding_factor
        # set new values in the last commitment
        last_commitment.update(new_last_commitment)
        blinding_outs[-1].value_blinding_factor = last_vbf

        # calculate commitments (surj proof etc)

        in_tags = [inp.asset for inp in pset.inputs]
        in_gens = [secp256k1.generator_parse(inp.utxo.asset) for inp in pset.inputs]

        for i, out in py_enumerate(pset.outputs):
            if out.blinding_pubkey is None:
                continue
            gen = secp256k1.generator_generate_blinded(
                out.asset, out.asset_blinding_factor
            )
            out.asset_commitment = secp256k1.generator_serialize(gen)
            value_commitment = secp256k1.pedersen_commit(
                out.value_blinding_factor, out.value, gen
            )
            out.value_commitment = secp256k1.pedersen_commitment_serialize(
                value_commitment
            )

            proof_seed = hashes.tagged_hash(
                "liquid/surjection_proof", txseed + i.to_bytes(4, "little")
            )
            proof, in_idx = secp256k1.surjectionproof_initialize(
                in_tags, out.asset, proof_seed
            )
            secp256k1.surjectionproof_generate(
                proof,
                in_idx,
                in_gens,
                gen,
                pset.inputs[in_idx].asset_blinding_factor,
                out.asset_blinding_factor,
            )
            out.surjection_proof = secp256k1.surjectionproof_serialize(proof)
            del proof

            # generate range proof
            rangeproof_nonce = hashes.tagged_hash(
                "liquid/range_proof", txseed + i.to_bytes(4, "little")
            )
            # reblind with extra message for unblinding
            out.reblind(
                rangeproof_nonce,
                extra_message=out.unknown.get(b"\xfc\x07specter\x01", b""),
            )

        return commitments

    def sign_pset(self, b64pset: str) -> str:
        """Signs specter-desktop specific Liquid PSET transaction"""
        mfp = self.get_master_fingerprint()
        pset = PSET.from_string(b64pset)
        commitments = self._blind(pset)

        ins = [
            {
                "is_witness": True,
                # "input_tx": inp.non_witness_utxo.serialize(),
                "script": inp.witness_script.data
                if inp.witness_script
                else script.p2pkh_from_p2wpkh(inp.script_pubkey).data,
                "value_commitment": inp.witness_utxo.value,
                "path": [
                    der
                    for der in inp.bip32_derivations.values()
                    if der.fingerprint == mfp
                ][0].derivation,
            }
            for inp in pset.inputs
        ]
        change = [
            {
                "path": [
                    der
                    for pub, der in out.bip32_derivations.items()
                    if der.fingerprint == mfp
                ][0].derivation,
                "variant": self._get_script_type(out),
            }
            if out.bip32_derivations and self._get_script_type(out) is not None
            else None
            for out in pset.outputs
        ]
        rawtx = pset.blinded_tx.serialize()

        signatures = self.jade.sign_liquid_tx(
            self._network(), rawtx, ins, commitments, change
        )
        for i, inp in py_enumerate(pset.inputs):
            inp.partial_sigs[
                [
                    pub
                    for pub, der in inp.bip32_derivations.items()
                    if der.fingerprint == mfp
                ][0]
            ] = signatures[i]
        # we must finalize here because it has different commitments and only supports singlesig
        return str(finalize_psbt(pset))

    def _get_script_type(self, out):
        if out.script_pubkey.script_type() == "p2pkh":
            return "pkh(k)"
        elif out.script_pubkey.script_type() == "p2wpkh":
            return "wpkh(k)"
        elif out.script_pubkey.script_type() == "p2sh":
            if out.redeem_script.script_type() == "p2wpkh":
                return "sh(wpkh(k))"
        # otherwise None
        return None


def enumerate(password=""):
    results = []

    # Jade is not really an HID device, it shows as a serial/com port device.
    # Scan com ports looking for the relevant vid and pid, and use 'path' to
    # hold the path to the serial port device, eg. /dev/ttyUSB0
    for devinfo in list_ports.comports():
        if devinfo.vid == JADE_VENDOR_ID and devinfo.pid == JADE_DEVICE_ID:
            d_data = {}
            d_data["type"] = "jade"
            d_data["path"] = devinfo.device
            d_data["needs_pin_sent"] = False
            d_data["needs_passphrase_sent"] = False

            client = None
            with handle_errors(common_err_msgs["enumerate"], d_data):
                client = JadeClient(devinfo.device, password)
                d_data["fingerprint"] = client.get_master_fingerprint().hex()

            if client:
                client.close()

            results.append(d_data)

    return results



import logging
import click
import os
from .cli_server import server
from .cli_noded import bitcoind, elementsd

logger = logging.getLogger(__name__)


@click.group()
@click.option("--debug", is_flag=True, help="Show debug information on errors.")
@click.option(
    "--tracerpc/--no-tracerpc",
    default=False,
    help="Will trace all calls to BitcoinCore or ElementsCore if in --debug",
)
@click.pass_context
def entry_point(config_home, debug=False, tracerpc=False):
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    if debug:
        # No need for timestamps while developing
        formatter = logging.Formatter("[%(levelname)7s] in %(module)15s: %(message)s")
        logging.getLogger("cryptoadvance").setLevel(logging.DEBUG)
    else:
        formatter = logging.Formatter(
            # Too early to format that via the flask-config, so let's copy it from there:
            os.getenv(
                "SPECTER_LOGFORMAT",
                "[%(asctime)s] %(levelname)s in %(module)s: %(message)s",
            )
        )
        logging.getLogger("cryptoadvance").setLevel(logging.INFO)
    ch.setFormatter(formatter)
    logging.getLogger().addHandler(ch)

    if tracerpc:
        if not debug:
            raise Exception("--tracerpc Doesn't make sense without --debug")
        logging.getLogger("cryptoadvance.specter.rpc").setLevel(logging.DEBUG)
    else:
        logging.getLogger("cryptoadvance.specter.rpc").setLevel(logging.INFO)


entry_point.add_command(server)
entry_point.add_command(bitcoind)
entry_point.add_command(elementsd)



from ..wallet import *
from ..addresslist import Address
from embit import ec
from embit.liquid.pset import PSET
from embit.liquid.transaction import LTransaction
from .txlist import LTxList
from .addresslist import LAddressList
from embit.liquid.addresses import to_unconfidential
from ..specter_error import SpecterError


class LWallet(Wallet):
    MIN_FEE_RATE = 0.1
    AddressListCls = LAddressList
    TxListCls = LTxList
    TxCls = LTransaction
    PSBTCls = PSET

    @classmethod
    def create(
        cls,
        rpc,
        rpc_path,
        working_folder,
        device_manager,
        wallet_manager,
        name,
        alias,
        sigs_required,
        key_type,
        keys,
        devices,
        core_version=None,
    ):
        """Creates a wallet. If core_version is not specified - get it from rpc"""
        # get xpubs in a form [fgp/der]xpub from all keys
        xpubs = [key.metadata["combined"] for key in keys]
        recv_keys = ["%s/0/*" % xpub for xpub in xpubs]
        change_keys = ["%s/1/*" % xpub for xpub in xpubs]
        is_multisig = len(keys) > 1
        # we start by constructing an argument for descriptor wrappers
        if is_multisig:
            recv_descriptor = "sortedmulti({},{})".format(
                sigs_required, ",".join(recv_keys)
            )
            change_descriptor = "sortedmulti({},{})".format(
                sigs_required, ",".join(change_keys)
            )
        else:
            recv_descriptor = recv_keys[0]
            change_descriptor = change_keys[0]
        # now we iterate over script-type in reverse order
        # to get sh(wpkh(xpub)) from sh-wpkh and xpub
        arr = key_type.split("-")
        for el in arr[::-1]:
            recv_descriptor = "%s(%s)" % (el, recv_descriptor)
            change_descriptor = "%s(%s)" % (el, change_descriptor)

        # get blinding key for the wallet
        blinding_key = None
        if len(devices) == 1:
            blinding_key = devices[0].blinding_key
            if not blinding_key:
                raise SpecterError(
                    "Device doesn't have a blinding key. Please import it."
                )

        # if we don't have slip77 key for a device or it is multisig
        # we use chaincodes to generate slip77 key.
        if not blinding_key:
            desc = LDescriptor.from_string(recv_descriptor)
            # For now we use sha256(b"blinding_key", xor(chaincodes)) as a blinding key
            # where chaincodes are corresponding to xpub of the first receiving address.
            # It's not a standard but we use that until musig(blinding_xpubs) is implemented.
            # Chaincodes of the first address are not used anywhere else so they can be used
            # as a source for the blinding keys. They are also independent of the xpub's origin.
            xor = bytearray(32)
            desc_keys = desc.derive(0).keys
            for k in desc_keys:
                if k.is_extended:
                    chaincode = k.key.chain_code
                    for i in range(32):
                        xor[i] = xor[i] ^ chaincode[i]
            secret = hashlib.sha256(b"blinding_key" + bytes(xor)).digest()
            blinding_key = ec.PrivateKey(secret).wif()
        if blinding_key:
            recv_descriptor = f"blinded(slip77({blinding_key}),{recv_descriptor})"
            change_descriptor = f"blinded(slip77({blinding_key}),{change_descriptor})"

        recv_descriptor = AddChecksum(recv_descriptor)
        change_descriptor = AddChecksum(change_descriptor)
        assert recv_descriptor != change_descriptor

        # get Core version if we don't know it
        if core_version is None:
            core_version = rpc.getnetworkinfo().get("version", 0)

        use_descriptors = core_version >= 209900
        # v20.99 is pre-v21 Elements Core for descriptors
        if use_descriptors:
            # Use descriptor wallet
            rpc.createwallet(os.path.join(rpc_path, alias), True, True, "", False, True)
        else:
            rpc.createwallet(os.path.join(rpc_path, alias), True)

        wallet_rpc = rpc.wallet(os.path.join(rpc_path, alias))
        # import descriptors
        args = [
            {
                "desc": desc,
                "internal": change,
                "timestamp": "now",
                "watchonly": True,
            }
            for (change, desc) in [(False, recv_descriptor), (True, change_descriptor)]
        ]
        for arg in args:
            if use_descriptors:
                arg["active"] = True
            else:
                arg["keypool"] = True
                arg["range"] = [0, cls.GAP_LIMIT]

        assert args[0] != args[1]

        # Descriptor wallets were introduced in v0.21.0, but upgraded nodes may
        # still have legacy wallets. Use getwalletinfo to check the wallet type.
        # The "keypool" for descriptor wallets is automatically refilled
        if use_descriptors:
            res = wallet_rpc.importdescriptors(args)
        else:
            res = wallet_rpc.importmulti(args, {"rescan": False})

        assert all([r["success"] for r in res])

        return cls(
            name,
            alias,
            "{} of {} {}".format(sigs_required, len(keys), purposes[key_type])
            if len(keys) > 1
            else purposes[key_type],
            addrtypes[key_type],
            "",
            -1,
            "",
            -1,
            0,
            0,
            recv_descriptor,
            change_descriptor,
            keys,
            devices,
            sigs_required,
            {},
            [],
            os.path.join(working_folder, "%s.json" % alias),
            device_manager,
            wallet_manager,
        )

    def getdata(self):
        self.fetch_transactions()
        self.check_utxo()
        self.get_info()
        # TODO: Should do the same for the non change address (?)
        # check if address was used already
        try:
            value_on_address = self.rpc.getreceivedbyaddress(
                self.change_address, assetlabel=None
            )
        except Exception as e:
            # Could happen if address not in wallet (wallet was imported)
            # try adding keypool
            logger.info(
                f"Didn't get transactions on change address {self.change_address}. Refilling keypool."
            )
            logger.error(e)
            self.keypoolrefill(0, end=self.keypool, change=False)
            self.keypoolrefill(0, end=self.change_keypool, change=True)
            value_on_address = {}

        # if not - just return
        if sum(value_on_address.values(), 0) > 0:
            self.change_index += 1
            self.getnewaddress(change=True)

    def get_balance(self):
        try:
            full_balance = (
                self.rpc.getbalances(assetlabel=None)["mine"]
                if self.use_descriptors
                else self.rpc.getbalances(assetlabel=None)["watchonly"]
            )
            balance = {"assets": {}}
            assets = {
                "bitcoin",
            }
            # get all assets
            for k in full_balance:
                for asset in full_balance[k].keys():
                    assets.add(asset)
            # get balances for every asset
            for asset in assets:
                asset_balance = {}
                for cat in full_balance:
                    v = full_balance[cat].get(asset, 0)
                    asset_balance[cat] = v
                if asset == "bitcoin":
                    balance.update(asset_balance)
                else:
                    balance["assets"][asset] = asset_balance

            # calculate available balance
            available = {}
            available.update(balance)
            # locked_utxo = self.rpc.listlockunspent()
            # we need better tx decoding here to include assets
            # for tx in locked_utxo:
            #     tx_data = self.gettransaction(tx["txid"])
            #     raw_tx = decoderawtransaction(tx_data["hex"], self.manager.chain)
            #     delta = raw_tx["vout"][tx["vout"]]["value"]
            #     if "confirmations" not in tx_data or tx_data["confirmations"] == 0:
            #         available["untrusted_pending"] -= delta
            #     else:
            #         available["trusted"] -= delta
            #         available["trusted"] = round(available["trusted"], 8)
            # available["untrusted_pending"] = round(available["untrusted_pending"], 8)
            balance["available"] = available
        except:
            balance = {
                "trusted": 0,
                "untrusted_pending": 0,
                "immature": 0,
                "available": {"trusted": 0, "untrusted_pending": 0},
                "assets": {},
            }
        self.balance = balance
        return self.balance

    def createpsbt(
        self,
        addresses: [str],
        amounts: [float],
        subtract: bool = False,
        subtract_from: int = 0,
        fee_rate: float = 1.0,
        selected_coins=[],
        readonly=False,
        rbf=True,
        existing_psbt=None,
        rbf_edit_mode=False,
        assets=None,
    ):
        """
        fee_rate: in sat/B or BTC/kB. If set to 0 Bitcoin Core sets feeRate automatically.
        """
        if fee_rate > 0 and fee_rate < self.MIN_FEE_RATE:
            fee_rate = self.MIN_FEE_RATE

        options = {"includeWatching": True, "replaceable": rbf}
        extra_inputs = []
        # get change addresses for all assets + for LBTC
        change_addresses = [
            self.get_address(self.change_index + i, change=True, check_keypool=False)
            for i in range(len(assets) + 1)
        ]

        if not existing_psbt:
            # if not rbf_edit_mode:
            #     if self.full_available_balance < sum(amounts):
            #         raise SpecterError(
            #             "The wallet does not have sufficient funds to make the transaction."
            #         )

            # if selected_coins != []:
            #     still_needed = sum(amounts)
            #     for coin in selected_coins:
            #         coin_txid = coin.split(",")[0]
            #         coin_vout = int(coin.split(",")[1])
            #         coin_amount = self.gettransaction(coin_txid, decode=True)["vout"][
            #             coin_vout
            #         ]["value"]
            #         extra_inputs.append({"txid": coin_txid, "vout": coin_vout})
            #         still_needed -= coin_amount
            #         if still_needed < 0:
            #             break
            #     if still_needed > 0:
            #         raise SpecterError(
            #             "Selected coins does not cover Full amount! Please select more coins!"
            #         )
            # elif self.available_balance["trusted"] <= sum(amounts):
            #     txlist = self.rpc.listunspent(0, 0)
            #     b = sum(amounts) - self.available_balance["trusted"]
            #     for tx in txlist:
            #         extra_inputs.append({"txid": tx["txid"], "vout": tx["vout"]})
            #         b -= tx["amount"]
            #         if b < 0:
            #             break

            # subtract fee from amount of this output:
            # currently only one address is supported, so either
            # empty array (subtract from change) or [0]
            subtract_arr = [subtract_from] if subtract else []

            options = {
                "includeWatching": True,
                # FIXME: get back change addresses
                # "changeAddress": self.change_address,
                "subtractFeeFromOutputs": subtract_arr,
                "replaceable": rbf,
                "changeAddresses": change_addresses,  # not supported by Elements - custom field for out LiquidRPC
            }

            # 209900 is pre-v21 for Elements Core
            if self.manager.bitcoin_core_version_raw >= 209900:
                options["add_inputs"] = selected_coins == []

            if fee_rate > 0:
                # bitcoin core needs us to convert sat/B to BTC/kB
                options["feeRate"] = round((fee_rate * 1000) / 1e8, 8)

            # looks like change_type is required in nested segwit wallets
            # but not in native segwit
            if "changeAddress" not in options and self.address_type:
                options["change_type"] = self.address_type

            r = self.rpc.walletcreatefundedpsbt(
                extra_inputs,  # inputs
                [
                    {addresses[i]: amounts[i], "asset": assets[i]}
                    for i in range(len(addresses))
                ],  # output
                0,  # locktime
                options,  # options
                True,  # bip32-der
            )

            b64psbt = r["psbt"]
            psbt = self.rpc.decodepsbt(b64psbt)
        else:
            psbt = existing_psbt
            # vins from psbt v0 or v2
            if "tx" in psbt:
                extra_inputs = [
                    {"txid": tx["txid"], "vout": tx["vout"]} for tx in psbt["tx"]["vin"]
                ]
            else:
                extra_inputs = [
                    {"txid": inp["previous_txid"], "vout": inp["previous_vout"]}
                    for inp in psbt["inputs"]
                ]
            # FIXME: get back change addresses
            # if "changeAddress" in psbt:
            #     options["changeAddress"] = psbt["changeAddress"]
            #     if "change_type" in options:
            #         options.pop("change_type")
            if "base64" in psbt:
                b64psbt = psbt["base64"]

        # looks like change_type is required in nested segwit wallets
        # but not in native segwit
        if "changeAddress" not in options and self.address_type:
            options["change_type"] = self.address_type

        if fee_rate > 0.0:
            if not existing_psbt:
                adjusted_fee_rate = self.adjust_fee(psbt, fee_rate)
                options["feeRate"] = "%.8f" % round((adjusted_fee_rate * 1000) / 1e8, 8)
            else:
                options["feeRate"] = "%.8f" % round((fee_rate * 1000) / 1e8, 8)
            r = self.rpc.walletcreatefundedpsbt(
                extra_inputs,  # inputs
                [
                    {addresses[i]: amounts[i], "asset": assets[i]}
                    for i in range(len(addresses))
                ],  # output
                0,  # locktime
                options,  # options
                True,  # bip32-der
            )

            b64psbt = r["psbt"]
            psbt = self.rpc.decodepsbt(b64psbt)
            psbt["fee_rate"] = options["feeRate"]
        # estimate full size
        tx_full_size = ceil(
            psbt["tx"]["vsize"] + len(psbt["inputs"]) * self.weight_per_input / 4
        )
        psbt["tx_full_size"] = tx_full_size

        psbt["base64"] = b64psbt
        psbt["amount"] = amounts
        psbt["address"] = addresses
        if assets:
            psbt["asset"] = assets
        psbt["time"] = time.time()
        psbt["sigs_count"] = 0
        if not readonly:
            self.save_pending_psbt(psbt)

        return psbt

    def adjust_fee(self, psbt, fee_rate):
        psbt_fees_sats = int(psbt.get("fees", {}).get("bitcoin", 0) * 1e8)

        # TODO: handle non-blind outputs differently
        num_blinded_outs = len(psbt["outputs"]) - 1
        # estimate final size: add weight of inputs and outputs
        # out witness weight is 4245 (from some random tx)
        # commitments in blinded tx: 33 for nonce and 33 for value
        commitment_size = 33 + 33
        tx_full_size = ceil(
            psbt["tx"]["vsize"]
            + len(psbt["inputs"]) * self.weight_per_input / 4
            + num_blinded_outs * 4245 / 4
            # probably elements doesn't count commitments
            + len(psbt["inputs"]) * commitment_size
            + num_blinded_outs * commitment_size
        )
        return (
            fee_rate
            * (fee_rate / (psbt_fees_sats / psbt["tx"]["vsize"]))
            * (tx_full_size / psbt["tx"]["vsize"])
        )

    def addresses_info(self, is_change):
        """Create a list of (receive or change) addresses from cache and retrieve the
        related UTXO and amount.
        Parameters: is_change: if true, return the change addresses else the receive ones.
        """

        addresses_info = []

        addresses_cache = [
            v for _, v in self._addresses.items() if v.change == is_change
        ]

        for addr in addresses_cache:

            addr_utxo = 0
            addr_amount = 0
            addr_assets = {}

            for utxo in [
                utxo
                for utxo in self.full_utxo
                if to_unconfidential(utxo["address"]) == to_unconfidential(addr.address)
            ]:
                addr_amount = addr_amount + utxo["amount"]
                addr_utxo = addr_utxo + 1
                addr_assets[utxo.get("asset")] = (
                    addr_assets.get(utxo.get("asset"), 0) + utxo["amount"]
                )

            addresses_info.append(
                {
                    "index": addr.index,
                    "address": addr.address,
                    "label": addr.label,
                    "amount": addr_amount,
                    "used": bool(addr.used),
                    "utxo": addr_utxo,
                    "type": "change" if is_change else "receive",
                    "assets": addr_assets,
                }
            )

        return addresses_info

    @property
    def unconfidential_address(self):
        return to_unconfidential(self.address)



### bc-ur encoding stuff
from io import BytesIO
import hashlib

CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def bech32_polymod(values):
    """Internal function that computes the Bech32 checksum."""
    generator = [0x3B6A57B2, 0x26508E6D, 0x1EA119FA, 0x3D4233DD, 0x2A1462B3]
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1FFFFFF) << 5 ^ value
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrp_expand(hrp):
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


def bech32_verify_checksum(hrp, data):
    """Verify a checksum given HRP and converted data characters."""
    return bech32_polymod(bech32_hrp_expand(hrp) + data) == 1


def bech32_create_checksum(hrp, data):
    """Compute the checksum values given HRP and data."""
    values = bech32_hrp_expand(hrp) + data
    polymod = bech32_polymod(values + [0, 0, 0, 0, 0, 0]) ^ 1
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def bech32_encode(hrp, data):
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data)
    return hrp + "1" + "".join([CHARSET[d] for d in combined])


def bech32_decode(bech):
    """Validate a Bech32 string, and determine HRP and data."""
    if (any(ord(x) < 33 or ord(x) > 126 for x in bech)) or (
        bech.lower() != bech and bech.upper() != bech
    ):
        return (None, None)
    bech = bech.lower()
    pos = bech.rfind("1")
    if pos < 1 or pos + 7 > len(bech) or len(bech) > 90:
        return (None, None)
    if not all(x in CHARSET for x in bech[pos + 1 :]):
        return (None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos + 1 :]]
    if not bech32_verify_checksum(hrp, data):
        return (None, None)
    return (hrp, data[:-6])


def convertbits(data, frombits, tobits, pad=True):
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            return None
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        return None
    return ret


def decode(hrp, addr):
    """Decode a segwit address."""
    hrpgot, data = bech32_decode(addr)
    if hrpgot != hrp:
        return (None, None)
    decoded = convertbits(data[1:], 5, 8, False)
    if decoded is None or len(decoded) < 2 or len(decoded) > 40:
        return (None, None)
    if data[0] > 16:
        return (None, None)
    if data[0] == 0 and len(decoded) != 20 and len(decoded) != 32:
        return (None, None)
    return (data[0], decoded)


def encode(hrp, witver, witprog):
    """Encode a segwit address."""
    ret = bech32_encode(hrp, [witver] + convertbits(witprog, 8, 5))
    if decode(hrp, ret) == (None, None):
        return None
    return ret


def bc32encode(data: bytes) -> str:
    """
    bc32 encoding
    see https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-004-bc32.md
    """
    dd = convertbits(data, 8, 5)
    polymod = bech32_polymod([0] + dd + [0, 0, 0, 0, 0, 0]) ^ 0x3FFFFFFF
    chk = [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]
    return "".join([CHARSET[d] for d in dd + chk])


def bc32decode(bc32: str) -> bytes:
    """
    bc32 decoding
    see https://github.com/BlockchainCommons/Research/blob/master/papers/bcr-2020-004-bc32.md
    """
    if bc32.lower() != bc32 and bc32.upper() != bc32:
        return None
    bc32 = bc32.lower()
    if not all([x in CHARSET for x in bc32]):
        return None
    res = [CHARSET.find(c) for c in bc32.lower()]
    if bech32_polymod([0] + res) != 0x3FFFFFFF:
        return None
    return bytes(convertbits(res[:-6], 5, 8, False))


def cbor_encode(data):
    l = len(data)
    if l <= 23:
        prefix = bytes([0x40 + l])
    elif l <= 255:
        prefix = bytes([0x58, l])
    elif l <= 65535:
        prefix = b"\x59" + l.to_bytes(2, "big")
    else:
        prefix = b"\x60" + l.to_bytes(4, "big")
    return prefix + data


def cbor_decode(data):
    s = BytesIO(data)
    b = s.read(1)[0]
    if b >= 0x40 and b < 0x58:
        l = b - 0x40
        return s.read(l)
    if b == 0x58:
        l = s.read(1)[0]
        return s.read(l)
    if b == 0x59:
        l = int.from_bytes(s.read(2), "big")
        return s.read(l)
    if b == 0x60:
        l = int.from_bytes(s.read(4), "big")
        return s.read(l)
    return None


def bcur_encode(data):
    """Returns bcur encoded string and hash digest"""
    cbor = cbor_encode(data)
    enc = bc32encode(cbor)
    h = hashlib.sha256(cbor).digest()
    enc_hash = bc32encode(h)
    return enc, enc_hash


def bcur_decode(data, checksum=None):
    """Returns decoded data, verifies hash digest if provided"""
    cbor = bc32decode(data)
    if checksum is not None:
        h = bc32decode(checksum)
        assert h == hashlib.sha256(cbor).digest()
    return cbor_decode(cbor)


if __name__ == "__main__":
    from binascii import a2b_base64

    b64 = "cHNidP8BAHEBAAAAAfPQ5Rpeu5nH0TImK4Sbu9lxIOGEynRadywPxaPyhnTwAAAAAAD/////AkoRAAAAAAAAFgAUFCYoQzGSRmYVAuZNuXF0OrPg9jWIEwAAAAAAABYAFOZMlwM1sZGLivwOcOh77amAlvD5AAAAAAABAR+tKAAAAAAAABYAFM4u9V5WG+Fe9l3MefmYEX4ULWAWIgYDA+jO+oOuN37ABK67BA/+SuuR/57c7OkyfyR7hR34FDsYccBxUlQAAIAAAACAAAAAgAAAAAAFAAAAACICApJMZBvzWiavLN7nievKQoylwPoffLkXZUIgGHF4HgwaGHHAcVJUAACAAAAAgAAAAIABAAAACwAAAAAA"
    raw = a2b_base64(b64)
    enc, enc_hash = bcur_encode(raw)
    dec = bcur_decode(enc, enc_hash)
    assert dec == raw
    testres = [
        "tyq3wurnvf607qgqwyqsqqqqq8eapeg6t6aen373xgnzhpymh0vhzg8psn98gknh9s8utgljse60qqqqqqqqpllllllsyjs3qqqqqqqqqqtqq9q5yc5yxvvjgenp2qhxfkuhzap6k0s0vdvgzvqqqqqqqqqpvqq5uexfwqe4kxgchzhupecws7ld4xqfdu8eqqqqqqqq",
        "qyq3ltfgqqqqqqqqqqtqq9xw9m64u4smu900vhwv08uesyt7zskkq93zqcps86xwl2p6udm7cqz2awcyplly46u3l70dem8fxfljg7u9rhupgwccw8q8z5j5qqqgqqqqqzqqqqqqsqqqqqqqq5qqqqqqygpq9yjvvsdlxk3x4ukdaeufa09y9r99crap7l9ezaj5ygqc",
        "w9upurq6rpcuqu2j2sqqpqqqqqqgqqqqqzqqzqqqqq9sqqqqqqqqmkdau4",
    ]
    testhash = "hlwjxjx550k4nnfdl5py2tn3vnh6g60slnw5dmld6ktrkkz200as49spg5"
    assert enc_hash == testhash
    assert enc == "".join(testres)
